/* GSI - general sound interface  [-*- C++ -*-]
 *
 * gsi_interface.c (non API dependent version, for the socket server)
 *
 * Copyright (C) W.H.Scholten 1996,1997,1998
 *
 * ** See gsi.h for a description of the commands **
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */

/* $Log: gsi_interface.c,v $
 * Revision 1.13  1998/05/28 09:37:23  wouter
 * small updates
 *
 * Revision 1.12  1998/05/26 19:54:40  wouter
 *
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdarg.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>

#include <gsi/config.h>
#define USE_SOCKETS

#define  _GSI_INTERFACE_
#include <gsi/gsi.h>

#include <gsi/mytypes.h> /* use gsi_int16 etc instead ? */
#include <gsi/debug.h>


#ifdef USE_SOCKETS
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netdb.h>
#endif

/*__________________VARIABELEN___________________*/
char *gsi_version=GSI_INTERFACE_VERSION;

static char *gsi_server;
static char *mixerdev=MIXER_DEV;
static int  sound_on=1;
//more? gsi_digi_on/gsi_midi_on/gsi_cd_on
//  OR: gsi_sound_on/gsi_music_on/gsi_cd_on

static int  soundfd;
static int  statusfd;

static char sndserver_ok = 0;

static int  num_commands;
static byte gsi_command_buffer[256];

static int  GSI_volume[4] = {-1,-1,-1,-1};


static int  server_id; /* process id of the gsi_server */

static int  channel_status[8]; /* 256 bits (playing yes/no for each channel), i.e. an int must be 32 bit or more */

/* Call back function jump table */
static void (*cb_function[1+256])(int channel); /* all channels and one for songs */

static void (*error_callback_function)(int error);

#ifdef USE_SOCKETS
int socket_fd;				/* connected socket descriptor */

struct hostent *hp;		/* pointer to host info for remote host */
struct servent *sp;		/* pointer to service information */

long timevar;			/* contains time returned by time() */

struct sockaddr_in myaddr_in;	/* for local socket address */
struct sockaddr_in peeraddr_in;	/* for peer socket address */
#endif /* USE_SOCKETS */

/*__________________DEFINITIES___________________*/
#define LINEAR_VOLUME

#define MAX_QUEUE_LENGTH 256 /* sound commands stored */

#define GSI_SERVER "gsi_server"

/*__________________PROTOTYPES___________________*/
static int  _gsi_get_channel_status (int ch);
static void _gsi_set_channel_status (int ch, int bit);
//Low level
void gsi_send_byte (int cmd);
//Direct to the pipe/socket:
void _gsi_send_byte (int k);
void _gsi_send_int16 (int k);
void _gsi_send_int32 (int k);
void _gsi_send_string (char *s);
void _gsi_send_block (byte *data, int len);

static int init_server ();


/*__________________PROGRAMMA___________________*/
#ifndef USE_SOCKETS
/* This code is only useful for the execvp'ed server */
static void child_killed(int i) {
	int status;

//	printf("child killed, i=%d\n", i);

	if (waitpid(server_id, &status, WNOHANG) == server_id) {
		sndserver_ok = 0;
		close(soundfd);
		close(statusfd);

		fprintf(stderr, "gsi_server killed\n");
	}
}
#endif /* !USE_SOCKETS */


/*---------------------------------------------------------------------------*
 *	Set sound volumes
 *  Mixer volume settings:
 *    VOLUME : master volume, i.e. headphone and line out.
 *    PCM    : dsp volume
 *    SYNTH  : onboard midi playing device
 *    CD     : cd output via the soundcard
 *
 *  saving volumes is done to remove any rounding problems with repeated
 *  reading/writing (device values read may not be the same as what has been
 *  set)
 *---------------------------------------------------------------------------*/
void gsi_set_master_volume(int vol) {
	gsi_send_commands(3, SND_SET_MIXER, GSI_MASTER_VOLUME, vol>0? vol-1: 0);
/* good idea to do this?:
 * GSI_volume[GSI_MASTER_VOLUME]=vol;
 * may be corrected with callback */
}
void gsi_set_pcm_volume(int vol) {
	gsi_send_commands(3, SND_SET_MIXER, GSI_PCM_VOLUME, vol>0? vol-1: 0);
}
void gsi_set_synth_volume(int vol) {
	gsi_send_commands(3, SND_SET_MIXER, GSI_SYNTH_VOLUME, vol>0? vol-1: 0);
}
void gsi_set_cd_volume(int vol) {
	gsi_send_commands(3, SND_SET_MIXER, GSI_CD_VOLUME, vol>0? vol-1: 0);
}

int gsi_get_master_volume() {
	return GSI_volume[GSI_MASTER_VOLUME];
}
int gsi_get_pcm_volume() {
	return GSI_volume[GSI_PCM_VOLUME];
}
int gsi_get_synth_volume() {
	return GSI_volume[GSI_SYNTH_VOLUME];
}
int gsi_get_cd_volume() {
	return GSI_volume[GSI_CD_VOLUME];
}



/*---------------------------------------------------------------------------*
 *	GSI initialisation
 *	here read the hardware master volume, pcm volume, synth volume cd volume
 *	Or do this with a different piece of code above? (init volumes at -1)
 *---------------------------------------------------------------------------*/
int gsi_init() {
	int mixer;

	/* Here connect to the gsi_server (or fork it if connection fails?) */
	init_server();

	/* after connection, the server will (should) automatically send the current
	 * hardware volumes so we don't need to do
	 *   gsi_send_commands(2, SND_GET_MIXER, GSI_SYNTH_VOLUME)
	 * etc, here. */

	return 0;
}

/* Currently the server is a shared one for both DSP and MIDI
 * Another approach using multiple servers/or threads might be in order.
 */
static int init_server() {
#ifndef USE_SOCKETS
	int  child; /* process id of the child process */
	int  fd[2];
	int  fd2[2];
	char *argv[3];

	if (sndserver_ok) return True;

	if (!(gsi_server=getenv("GSI_SERVER"))) gsi_server=GSI_SERVER;
	signal(SIGCHLD, &child_killed);

	/* should set SIGPIPE handler too */
//	signal(SIGPIPE, SIG_IGN);

	// Create a pipe to send data and commands:
	pipe(fd);
	soundfd=fd[1]; /* writeable pipe */

	// Create a pipe for status and callbacks:
	pipe(fd2);
	statusfd=fd2[0]; /* readable pipe */

	// Now set the write end to close when we exec the sound server,
	// and set both ends to non-blocking (fcntl below)


	sndserver_ok=True; /* preemptive, gets nuked on an error */

	child=fork();
	if( child == 0 )  { /* child's thread */
		close(fd[1]);
		close(fd2[0]);

		dup2(fd[0], STDIN_FILENO);
		close(fd[0]);
		dup2(fd2[1], STDOUT_FILENO);
		close(fd2[1]);

		argv[0] = gsi_server;
		argv[1] = "-loop";
		argv[2] = NULL;
		execvp(gsi_server, argv);
			// We only arrive here if execvp failed to replace this process
			// with the soundserver code
		fprintf (stderr, "GSI: Couldn't execute %s!\n", gsi_server);
		exit (0); /* child quits */
	}

	// Close unused ends:
	close(fd[0]);
	close(fd2[1]);

	server_id=child;
	/* should now wait for ok from the read-pipe */

	/* Call the plumber to unblock the 'in' pipe! */
	fcntl(fd2[0], F_SETFL, O_NONBLOCK);

	DEB1(fprintf(stderr, "GSI: sndserver_ok=%d\n", sndserver_ok));
	return sndserver_ok;
#else /* USE_SOCKETS */
	/* Connect to the server via a port */
	/* ==================================================================== */
	/*	We request a service called "gsi_server".
	 *	An entry for it needs to exist in the /etc/services file.
	 *	The port address for this service can be any port number that is
	 *	likely to be unused, such as 22375, for example.  The host on which
	 *	the server runs must also have the same entry (same port number) in
	 *	its /etc/services file. */

	int  addrlen, i, j;
	char *remote_host="localhost";

    int  curstate = 1;


	if (sndserver_ok) return True; /* already initialised */

	if (!(gsi_server=getenv("GSI_SERVER"))) gsi_server=GSI_SERVER;

	sndserver_ok=True; /* preemptive, gets nuked on an error */

	/* clear out address structures */
	memset ((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
	memset ((char *)&peeraddr_in, 0, sizeof(struct sockaddr_in));

	/* Set up the peer address to which we will connect. */
	peeraddr_in.sin_family = AF_INET;

	/* Get the host information for the hostname that the user passed in. */
	hp = gethostbyname (remote_host);
	if (hp == NULL) {
		fprintf(stderr, "%s: %s not found in /etc/hosts\n",
				gsi_version, remote_host);
		sndserver_ok=False;
		return -1;
//		exit(1);
	}
	peeraddr_in.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
	/* Find the information for the gsi_server in order to get the needed
	 * port number. */
	sp = getservbyname (gsi_server, "tcp");
	if (sp == NULL) {
		fprintf(stderr, "%s: gsi_server not found in /etc/services\n",
				gsi_version);
		sndserver_ok=False;
		return -1;
//		exit(1);
	}
	peeraddr_in.sin_port = sp->s_port;

	/* Create the socket. */
	socket_fd = socket (AF_INET, SOCK_STREAM, 0);
	if (socket_fd == -1) {
		perror(gsi_version);
		fprintf(stderr, "%s: unable to create socket\n", gsi_version);
		sndserver_ok=False;
		return -1;
//		exit(1);
	}

#if 1
	/* A few options: */
	/* Unblock the socket (now the electrician is busy) */
    if ( fcntl( socket_fd, F_SETFD, FD_CLOEXEC ) < 0 )	{
		fprintf(stderr,"Unable to set socket to 'close on exec'\n");
		return( -1 );
	}
	/* enable local address reuse */
    if ( setsockopt( socket_fd, SOL_SOCKET, SO_REUSEADDR,
					 &curstate, sizeof(curstate) ) < 0 ) {
		fprintf(stderr,"Unable to set for a fresh socket\n");
		return( -1 );
	}
	/* ------------- */
#endif


//	fcntl(socket_fd, F_SETFL, O_NONBLOCK);

	/* Try to connect to the remote server at the address which was just built
	 * into peeraddr. */
	if (connect(socket_fd, (struct sockaddr *) &peeraddr_in, sizeof(struct sockaddr_in)) == -1) {
		perror(gsi_version);
		fprintf(stderr, "%s: unable to connect to remote\n", gsi_version);
		sndserver_ok=False;
		return -1;
//		exit(1);
	}

	/* The connect call assigns a random address to the local end of the
	 * connection, so we use getsockname to see what it assigned. */
	addrlen = sizeof(struct sockaddr_in);
	if (getsockname(socket_fd, (struct sockaddr *) &myaddr_in, &addrlen) == -1) {
		perror(gsi_version);
		fprintf(stderr, "%s: unable to read socket address\n", gsi_version);
		sndserver_ok=False;
		return -1;
//		exit(1);
	}

	time(&timevar);
	fprintf(stderr, "Connected to %s on port %u at %s",
		   remote_host, ntohs(myaddr_in.sin_port), ctime(&timevar));

	fprintf(stderr, "gsi_interface: socket init done (socket no=%d)\n", socket_fd);
	fcntl(socket_fd, F_SETFL, O_NONBLOCK);

	soundfd=socket_fd;
	statusfd=socket_fd;

	sndserver_ok=True;
	return 0;
	/* ==================================================================== */
#endif /* USE_SOCKETS n/y */
}


/*---------------------------------------------------------------------------*
 *	Init PCM system, returns 0 if ok, -1 if failed.
 *	Note that this function may be called multiple times (it's just a
 *	convenient way to set a lot of stuff at once)
 *---------------------------------------------------------------------------*/
int  gsi_init_pcm(int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile) {
	init_server();

	if (sndserver_ok) {
			//server starts PAUSE'd, DSP released.
		_gsi_send_byte(SND_SET_DSP_SPEED);   _gsi_send_int16(dsp_speed);
		DEB1(fprintf(stderr, "GSI: set speed %d\n", dsp_speed));
		_gsi_send_byte(SND_SET_DSP_CHANNELS);_gsi_send_byte(dsp_channels);
		DEB1(fprintf(stderr, "GSI: set chan  %d\n", dsp_channels));
		_gsi_send_byte(SND_SET_DSP_BYTES);   _gsi_send_byte(dsp_bytes);
		DEB1(fprintf(stderr, "GSI: set bytes %d\n", dsp_bytes));
		_gsi_send_byte(SND_SET_DSP_FRAGSIZE);_gsi_send_byte(dsp_fragsize);
		DEB1(fprintf(stderr, "GSI: set frag  %d\n", dsp_fragsize));
		_gsi_send_byte(SND_SET_SOUND_DIR);   _gsi_send_string(sounddir);
		DEB1(fprintf(stderr, "GSI: set sounddir  %s\n", sounddir));
		_gsi_send_byte(SND_LOAD_SOUND_LIST); _gsi_send_string(soundfile);
		DEB1(fprintf(stderr, "GSI: set soundsfile  %s\n", soundfile));
		_gsi_send_byte(SND_LOAD_LISTED);
		DEB1(fprintf(stderr, "GSI: SND_LOAD_LISTED\n"));
		_gsi_send_byte(SND_GRAB_DSP); //This sets the new parameters.
		DEB1(fprintf(stderr, "GSI: SND_GRAB_DSP\n"));
		DEB1(fprintf(stderr, "%s initialized.\n", gsi_server));
		return 0;
	}
	else return -1;
}


/*---------------------------------------------------------------------------*
 *	Init SYNTH system, returns 0 if ok, -1 if failed.
 *---------------------------------------------------------------------------*/
int gsi_init_synth (int preferred_type) {
	init_server();

	if (sndserver_ok) {
//		gsi_send_commands(1, SND_GRAB_SYNTH);
		return 0;
	}
	return -1;
}


/*---------------------------------------------------------------------------*
 *	Play sounds 'n stuff
 *---------------------------------------------------------------------------*/
void gsi_kill_sounds() {
	num_commands=0;
	_gsi_send_byte(SND_KILLALL);
}


// Send a list of commands with data (bytes):
void gsi_send_commands(int argc, ...) {
	va_list argp;

	va_start(argp, argc);
	while (argc--) {
		gsi_command_buffer[num_commands++]=va_arg(argp, byte);
		while (num_commands>=MAX_QUEUE_LENGTH) gsi_flush();
	}
}


/* Sound playing ... */
void gsi_play_sound (int slot, int command) {
	if (sound_on /*&& sndserver_ok*/) {
		while (num_commands>=MAX_QUEUE_LENGTH-1) gsi_flush();
		DEB1(fprintf(stderr, "GSI: play_sound (slot %d, cmd %d), num_commands now %d\n", slot, command, num_commands+2));

		gsi_command_buffer[num_commands++]=command;
		gsi_command_buffer[num_commands++]=slot;
	}
}


void gsi_play_channel_sound (int channel, int slot, int command) {
	if (sound_on /*&& sndserver_ok*/) {
		while (num_commands>=MAX_QUEUE_LENGTH-2) gsi_flush();
		gsi_command_buffer[num_commands++]=command | SND_CHANNEL;
		gsi_command_buffer[num_commands++]=channel;
		gsi_command_buffer[num_commands++]=slot;
	}
}


void gsi_play_sound_3d (int ch, int slot, int command, int x, int y, int z, int dx, int dy, int dz, int doppler_on) {
	int  channel_flag=0;
	ASSERT(ch>=-1 && ch <256);

	if (ch!=-1) channel_flag=SND_CHANNEL;
	if (sound_on /*&& sndserver_ok*/) {
		gsi_send_commands(1, command | channel_flag);
		if (ch!=-1) gsi_send_commands(1, ch);
		gsi_send_commands(19, slot,
						  SND_SET_X  | SND_CURRENT | channel_flag, x >>8, x &255,
						  SND_SET_Y  | SND_CURRENT | channel_flag, y >>8, y &255,
						  SND_SET_Z  | SND_CURRENT | channel_flag, z >>8, z &255,
						  SND_SET_DX | SND_CURRENT | channel_flag, dx >>8, dx &255,
						  SND_SET_DY | SND_CURRENT | channel_flag, dy >>8, dy &255,
						  SND_SET_DZ | SND_CURRENT | channel_flag, dz >>8, dz &255
						  );
		if (doppler_on) gsi_send_commands(1, SND_DOPPLER_ON | channel_flag | SND_CURRENT);
	}
}


void gsi_play_sound_stereo (int ch, int slot, int command, int x, int dx, int doppler_on) {
	int  channel_flag=0;
	ASSERT(ch>=-1 && ch <256);

	if (ch!=-1) channel_flag=SND_CHANNEL;
	if (sound_on /*&& sndserver_ok*/) {
		gsi_send_commands(1, command | channel_flag);
		if (ch!=-1) gsi_send_commands(1, ch);
		gsi_send_commands(7, slot,
						  SND_SET_X  | SND_CURRENT | channel_flag, x >>8, x &255,
						  SND_SET_DX | SND_CURRENT | channel_flag, dx >>8, dx &255
						  );
		if (doppler_on) gsi_send_commands(1, SND_DOPPLER_ON | channel_flag | SND_CURRENT);
	}
}


void gsi_play_sound_pan (int ch, int slot, int command, int pan) {
	int  channel_flag=0;
	ASSERT(ch>=-1 && ch <256);
	ASSERT(pan <=256 && pan >=-256);

//this code could be made simple when using 255=NOP (default, ch_flag=255)
	if (ch!=-1) channel_flag=SND_CHANNEL;
	if (sound_on /*&& sndserver_ok*/) {
		gsi_send_commands(1, command | channel_flag);
		if (ch!=-1) gsi_send_commands(1, ch);
		gsi_send_commands(4, slot,
						  SND_SET_PAN  | SND_CURRENT | channel_flag, pan >>8, pan &255
						  );
	}
}


#if 0 /* unfinished , sends MSB first */
void gsi_play (struct gsi_sound_item *s, int effect_bits) {
	int  channel_flag=0;
	int  command=SND_PLAY;

	if (s->channel!=-1) channel_flag=SND_CHANNEL;
	if (s->repeat) command |=SND_REPEAT;
 
	if (sound_on /*&& sndserver_ok*/) {
		gsi_send_commands(1, command | channel_flag);
		if (ch!=-1) gsi_send_commands(1, ch);
/*		gsi_update(s, effect_bits);*/
		if (effect_bits & SND_BIT_PAN)
			gsi_send_commands(4, slot, SND_SET_PAN | SND_CURRENT |
							  channel_flag, s->pan >>8, s->pan &255
				);
		if (effect_bits & SND_BIT_X)
			gsi_send_commands(4, slot,SND_SET_X | SND_CURRENT |
							  channel_flag, s->x >>8, s->x &255
				);
		if (effect_bits & SND_ADJUST_X)
			gsi_send_commands(4, slot, SND_ADJUST_X | SND_CURRENT |
							  channel_flag, s->x >>8, s->x &255
				);
		if (effect_bits & SND_BIT_Z)
			gsi_send_commands(4, slot, SND_SET_Z | SND_CURRENT |
							  channel_flag, s->z >>8, s->z &255
				);
		if (effect_bits & SND_ADJUST_Z)
			gsi_send_commands(4, slot, SND_ADJUST_Z | SND_CURRENT |
							  channel_flag, s->z >>8, s->z &255
				);
		if (effect_bits & SND_BIT_Y)
			gsi_send_commands(4, slot, SND_SET_Y | SND_CURRENT |
							  channel_flag, s->y >>8, s->y &255
				);
		if (effect_bits & SND_ADJUST_Y)
			gsi_send_commands(4, slot, SND_ADJUST_Y | SND_CURRENT |
							  channel_flag, s->y >>8, s->y &255
				);

		if (effect_bits & SND_BIT_DX)
			gsi_send_commands(4, slot, SND_SET_DX| SND_CURRENT |
							  channel_flag, s->dx >>8, s->dx &255
				);
		if (effect_bits & SND_BIT_DZ)
			gsi_send_commands(4, slot, SND_SET_DZ  | SND_CURRENT |
							  channel_flag, s->dz >>8, s->dz &255
				);
		if (effect_bits & SND_BIT_DY)
			gsi_send_commands(4, slot, SND_SET_DY  | SND_CURRENT |
							  channel_flag, s->dy >>8, s->dy &255
				);
	}
}
#endif


/*---------------------------------------------------------------------------*
 *	Play music
 *---------------------------------------------------------------------------*/
int gsi_play_song(int loopno) {
	ASSERT(loopno>=0 && loopno<256);
	gsi_send_commands(2, SND_PLAY_SONG, loopno);
	return 0;
}
int gsi_stop_song() {
	gsi_send_commands(1, SND_STOP_SONG);
	return 0;
}
int gsi_continue_song() {
	gsi_send_commands(1, SND_CONTINUE_SONG);
	return 0;
}


/*---------------------------------------------------------------------------*
 *	Callbacks
 *---------------------------------------------------------------------------*/
static byte status_queue[4]; /* should use a larger circular buffer */
static int  status_queue_len;
void _gsi_check_callback () {
	byte b;
	int  i;
	int  decoded_length=0;

	DEB1(fprintf(stderr, "GSI: callback check\n"));
	if (sndserver_ok)
	for(;;) {
		i=read(statusfd, &b, 1);
		DEB1(fprintf(stderr, " -> read status=%d\n", i));
		if(i==0) break; //EOF on pipe, i.e. child died. Just quit and let the handler disable sound.
		DEB1(if (i<0) fprintf(stderr, "GSI: callback errno=%d\n", errno));
		if (i<0) { if (errno==EPIPE) sndserver_ok=0; break; /*nothing*/ }
		if (i>0) { //there was something
			DEB1(fprintf(stderr, "GSI: callback\n"));
			status_queue[status_queue_len++]=b;
//			int command=b; //=> if (command_size==0) command=b;
			switch(status_queue[0]) {
			case GSI_CALLBACK_OK:
				decoded_length=1;
				break;
			case GSI_CALLBACK_ERROR: /* general error */
				decoded_length=1;
				break;
			case GSI_CALLBACK_NO_SUCH_FILE:
				decoded_length=1;
					/* Should also send & get the slot number !! */
				DEB1(fprintf(stderr, "no such file\n"));
				if (error_callback_function!=NULL)
					error_callback_function(GSI_CALLBACK_NO_SUCH_FILE);
				break;
			case GSI_CALLBACK_BAD_LIST: /* eg sounds that can't be loaded ?? */
			case GSI_CALLBACK_BAD_COMMAND:
			case GSI_CALLBACK_DSP_BUSY:
			case GSI_CALLBACK_SYNTH_BUSY:
				decoded_length=1;
				break;
//			case GSI_CALLBACK_END: /* useless for the stack, unless we add sample id's generated when starting a sound, but must use callbacks for that too. Not worth it I think */
//				break;
			case GSI_CALLBACK_CHANNEL_END:
			{
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] &=~(1<<(j &31));
				DEB1(fprintf(stderr, "GSI_CALLBACK_CHANNEL_END: ch=%d\n", j));
				if (cb_function[j+1]!=NULL) (cb_function[j+1])(j);
			}
			break;
/*
  Why   is this   needed?  Suppose   a  sound  is   killed  or  ends,  a
  play_channel_sound is  given and gsi_set_channel_status(ch,  1) given,
  then a kill CALLBACK can come after this  and channel status will be 0
  even though the sound  has just started.   This means we need either a
  counter (increase with gsi_adjust_channel_status(ch) which increses it
  by 1, CALLBACK  decreases it by 1, or  callbacks (which will arrive in
  the same  order as things change in   the server) KILLALL must  give a
  callback to reset all counters to 0 
 */
			case GSI_CALLBACK_CHANNEL_STARTED:
			{
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] |=(1<<(j&31));
				DEB1(fprintf(stderr, "GSI_CALLBACK_CHANNEL_STARTED: ch=%d\n", j));
			}
			break;
			case GSI_CALLBACK_CHANNEL_ENDALL:
				decoded_length=1;
			{
				int  j;
				for (j=0; j<256; j++) {
					if (_gsi_get_channel_status(j) && cb_function[j+1]!=NULL)
						(cb_function[j+1])(j);
					_gsi_set_channel_status(j, 0);
				}
//				for (j=0; j<8; j++) channel_status[j] = 0;
				DEB1(fprintf(stderr, "GSI_CALLBACK_CHANNEL_ENDALL\n"));
			}
			break;
//			case GSI_CALLBACK_SONG_START:
//needed ?
			case GSI_CALLBACK_SONG_END:
			{
				decoded_length=1;
				DEB1(fprintf(stderr, "GSI_CALLBACK_SONG_END\n"));
				if (cb_function[0]!=NULL) (cb_function[0])(-1);
			}
			break;
			case GSI_STATUS_MIXER:
			{
				if (status_queue_len<3) continue;
				decoded_length=3;
				GSI_volume[status_queue[1]]=status_queue[2];
			}
			break;
			default: fprintf(stderr, "GSI: Unkown callback %d\n", status_queue[0]);
				decoded_length=1; /* remove the byte */
			}

			if (decoded_length>0) {
				status_queue_len -=decoded_length;
				if (status_queue_len>0) memmove(status_queue, status_queue+decoded_length, status_queue_len);
			}
		}
	}
}


void *gsi_set_channel_callback (int ch, void (*cbf)(int ch)) {
	void *old_func=cb_function[ch+1];
	ASSERT(ch>=-1 && ch <256);
	cb_function[ch+1]=cbf;

	return old_func;
}
void gsi_clear_channel_callback (int ch) {
	ASSERT(ch>=-1 && ch <256);
	cb_function[ch+1]=NULL;
}

void *gsi_set_error_callback (void (*cbf)(int error)) {
	void *old_func=error_callback_function;
	error_callback_function=cbf;

	return old_func;
}
void gsi_clear_error_callback () {
	error_callback_function=NULL;
}

int gsi_callback_fd () {
	return statusfd;
}


static int  _gsi_get_channel_status (int ch) {
	ASSERT(ch>=-1 && ch <256);
	return channel_status[(ch) >>5] & (1<< ((ch)&31));
}
static void _gsi_set_channel_status (int ch, int bit) {
	ASSERT(ch>=-1 && ch <256);
	channel_status[ch >>5] = (channel_status[ch >>5] & ~(1<< (ch &31))) | ( (bit &1) << (ch &31) );
}

int  gsi_get_channel_status (int ch) {
	_gsi_check_callback(); // handle any data on the input pipe
	return channel_status[ch >>5] & (1<< (ch&31));
}


/*---------------------------------------------------------------------------*
 *	Misc stuff
 *---------------------------------------------------------------------------*/
void gsi_close () {
	gsi_release_pcm(); //also release_cd/release_synth
	gsi_flush();
//	if (sndserver_ok) {
		_gsi_send_byte(SND_CLOSE_SERVER);
//	}
#if 0
	/* send EOF, not needed when we do 'close(soundfd)' */
	if (shutdown(soundfd, /*1*/ 2) == -1) {
		perror("shutdown");
		fprintf(stderr, "%s: unable to shutdown socket\n", gsi_version);
	}
#endif
	close(soundfd);

	sndserver_ok=0;
}


void gsi_flush () {
	if (num_commands>0)
		_gsi_send_block(&gsi_command_buffer[0], num_commands);
	num_commands=0;
	_gsi_check_callback();
}

void gsi_play_sound_flush (int slot, int command) {
	gsi_play_sound (slot, command);
	gsi_flush();
}

//separate pause_pcm?
//pause all (pcm/synth/cd)
void gsi_pause () {
//	if (sndserver_ok)
		gsi_send_byte(SND_PAUSE_ALL);
	gsi_flush();
}

//continue all (pcm/synth/cd)
void gsi_continue () {
//	if (sndserver_ok)
		gsi_send_byte(SND_CONTINUE_ALL);
	gsi_flush();
}

//add gsi_grab_devices() ?
void gsi_grab_pcm () {
//	if (sndserver_ok)
		gsi_send_byte(SND_GRAB_DSP);
	gsi_flush();
}


//add gsi_release_devices() ?
void gsi_release_pcm () {
//	if (sndserver_ok)
		gsi_send_byte(SND_RELEASE_DSP);
	gsi_flush();
}


//Data & file stuff
int  gsi_load_file (int slot, char *filename)
{
	gsi_flush(); //must do this as we access the pipe directly.

	_gsi_send_byte(SND_LOAD_SOUND_FILE);
	_gsi_send_byte(slot);
	_gsi_send_string(filename);
	return 0;
}
int  gsi_load_song_file (char *filename)
{
	gsi_flush(); //must do this as we access the pipe directly. ACTUALLY: HARD FLUSH NEEDED!

	_gsi_send_byte(SND_LOAD_SONG_FILE);
	_gsi_send_string(filename);
	return 0;
}

void gsi_send_sample (int slot, int speed, int type, int channels, int size, unsigned char *data) {
	gsi_flush(); //must do this as we access the pipe directly.

	/* This gets sent in sync mode */
	_gsi_send_byte(SND_LOAD_SOUND_DATA);
	_gsi_send_byte(slot);
	_gsi_send_int32(speed);
	_gsi_send_byte(type);
	_gsi_send_byte(channels);
	_gsi_send_int32(size);
	_gsi_send_block(data, size);
}


void gsi_send_byte(int cmd) {
	while (num_commands>=MAX_QUEUE_LENGTH) gsi_flush();
	gsi_command_buffer[num_commands++]=cmd;
}



/*---------------------------------------------------------------------------*
 *	Pipe/socket functions (internal use only)
 *---------------------------------------------------------------------------*/
void _gsi_send_block (byte *data, int len) {
	fd_set write_fd_sel;
	int i=0, j=0;

	DEB1(fprintf(stderr, "GSI: _gsi_send_block: sending a block of %d bytes\n", len));
	if (sndserver_ok) {
		while((j=write (soundfd, data+i, len-i)) <len-i) {
			if (j>0) i +=j; else if (errno==EPIPE) { DEB1(fprintf(stderr, "server dead\n")); sndserver_ok=0; break; }
			FD_ZERO(&write_fd_sel);
			FD_SET(soundfd, &write_fd_sel);
			select(soundfd+1, NULL, &write_fd_sel, NULL, NULL);
		}
	}
	DEB1(fprintf(stderr, "GSI: _gsi_send_block done\n"));
}
void _gsi_send_byte (int data) {
	fd_set write_fd_sel;
	byte pipe_byte=data &255;
	DEB2(fprintf(stderr, "GSI: _gsi_send_byte: sending %d\n", pipe_byte));
	if (sndserver_ok) {
		while(write (soundfd, &pipe_byte, sizeof (byte)) <1 && errno!=EPIPE) {
			FD_ZERO(&write_fd_sel);
			FD_SET(soundfd, &write_fd_sel);
			select(soundfd+1, NULL, &write_fd_sel, NULL, NULL);
		}
	}
}
void _gsi_send_int16 (int data) {
//	if (sndserver_ok) {
		_gsi_send_byte(data >>8);
		_gsi_send_byte(data);
//	}
}
void _gsi_send_int32 (int data) {
//	if (sndserver_ok) {
		_gsi_send_byte(data >>24);
		_gsi_send_byte(data >>16);
		_gsi_send_byte(data >>8);
		_gsi_send_byte(data);
//	}
}
void _gsi_send_string (char *string) {
//	if (sndserver_ok) {
		if (string==NULL)
			_gsi_send_byte(0);
		else
			_gsi_send_block(string, strlen(string)+1);
//	}
	DEB1(fprintf(stderr, "GSI: string length=%d\n", string==NULL ? 0 : strlen(string)));
}

/*---------------------------------------------------------------------------*
 *	EOF gsi_interface.c
 *---------------------------------------------------------------------------*/
