/* GSI - general sound interface   [-*- C++ -*-]
 *
 * gsi_server.c - shared sound server, main loop
 *   - sept 1996/may 1997/dec 1997/march-may 1998
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 *
 * Lower level API dependencies:
 *   - DSP mode setting: OSS but should not be too difficult to adapt
 *     to other systems.
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - sequencer/synth (MIDI's) playing: very OSS specific.
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */

/* socket version: call 'gsi_server -loop &' to let it run in the background */

/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <math.h>

#include <gsi/config.h>
#ifdef USE_OSS
#ifdef __linux__
#  include <sys/soundcard.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#  include <sys/errno.h>
#endif
#endif
#ifdef USE_NEW_BSD
#  error code for new bsd driver not implemented
#endif
#ifdef USE_DEV_AUDIO
#  warning some includes for /dev/audio may be missing
#endif

#include <gsi/mytypes.h>
#include <gsi/debug.h>
#define _SNDSERVER_
#include <gsi/gsi_commands.h>
#include <gsi/gsi_server_version.h>
#include "gsi_server.h"
#include "formats.h"
#include "mulaw.h"
#include "midi.h"


#define USE_SOCKETS



/*__________________DEFINITIES___________________*/
#ifdef DEBUG
#  define SEGFAULT_HANDLER
#endif


#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))


/*__________________EXTERN___________________*/
extern int  load_song(char *filename);
extern int  play_midi_init();
extern byte *midifilebuf;
extern int  playsong_ct;
extern int  playsong;
extern double current_time; /* midi timing */

#ifdef USE_OSS
SEQ_USE_EXTBUF();
#endif

// Client stuff
extern int  ls;				/* listen socket descriptor */


/*__________________VARIABELEN___________________*/
char *server_name;
char *dspdev;
/* ** DSP settings (init's are only used for set_sound_conversion_of(.) as that also gets called before the DSP is initialised. CHANGE THAT?? ) */
int  dsp_speed=8000;
int  dsp_type=GSI_8BIT;
int  dsp_log2_frag=8;
int  dsp_fragsize=256; /* samples=fragsize/channels/bytes */
int  dsp_channels=1;
int  dsp_bps; /* bytes per sample, 1 or 2 (determined from dsp_type) */

int  dsp_samples; // number of samples that fit in the fragsize


int  autorelease_dsp=False;
int  autoset_dsp=False;

/* ** various */
int  sinus_tabel[1024];  // for wave playing and observer rotation

int playnum; /* temp */
int channel_playnum; /* temp */
int are_using_dsp=False;


extern int  num_clients;
extern struct client_data *client_list[64];

int  current_client;
struct client_data *current;

int  dsp_request[64]; /* numbers of the client */
int  dsp_requests; /* number of clients that currently want the use of the DSP */

// Internal sound flags for actions on a sound:
#define SF_REPEAT            (1<<0)
#define SF_DOPPLER           (1<<1)
// Positional placement after sound changes:
#define SF_RECALC_3D_ONCE    (1<<2) /* eg. when x is set with SND_SET */
#define SF_RECALC_3D         (1<<3) /* when either dx,dy,dz are not zero */
#define SF_3D                (1<<4) //Necessary?? 3d if set, pan otherwise
#define SF_INIT_POSITION     (1<<5) /* To init radius[i] etc. */

#define SF_ECHO              (1<<6)
#define SF_OBSERVER_POS      (1<<7) /* alter sound positions/speeds according to the observer pos and rotation, i.e. all outside sounds, not e.g. sounds inside a car or game error beeps and stuff */


/* ** The sound mixer buffers */
/* should do checking below when a frag/bps/channel change is requested */
int  premix [1024*2];  /* n_samples */
byte final [1024*8]; /* n_samples*n_channels*n_bytes, i.e. <= n_samples*4 */

/* How best to alloc/realloc upon bytes/channel/fragsize change ? */
//byte *final;
//int  *premix;


// File descriptors:
int  dsp=-1;

int  command_fd, status_fd; /* The current ones we're working with */
/* For PIPE: command_fd=0 =STDIN_FILENO, status_fd=1=STDOUT_FILENO */
/* For SOCKET: command_fd=socket_fd, status_fd=socket_fd */


/*__________________FUNCTIE PROTOTYPES___________________*/
void  snd_server();
void  play_fragment(int slot);

void  init_sample(int slot, int speed, int type, int channels, int size);

void  set_sound_conversion(); /* all slots */
void  set_sound_conversion_of(int slot);

void  status_send_byte(int b);
void  status_flush();

void  make_sinus();

void  check_dxyz(int j); /* Sound flags wrt to 3d and doppler etc. */

int16 msb_int16(int fd); /* read MSB 16 bit int from file */
int32 msb_int32(int fd); /* read MSB 32 bit int from file */

void  get_block(byte *data, int len);

byte  get_byte();   // rename read_pipe_byte/read_byte_from_pipe ?
int16 get_int16();
int32 get_int32();

int32 swap_int32(int a);
int16 swap_int16(int16 a);

char *get_string(); // Read a string from pipe (for which storage is malloc'ed)

void  quit (int i);
void  segfault (int msg);

void  decode_command (int command);
void  remove_dsp_request(int client);


/* clients.c: */
extern int   init_connection();
extern int   add_client();
extern int   remove_client(int client);

/* dsp.c: */
extern int   setup_dsp(char *dspdev, int speed, int fragsize, int channels, int type);

/* mixer.c: */
extern int   init_mixer();
extern int   set_mixer(int dev, int vol);
extern int   read_mixer(int dev);

/* load.c: */
extern int   init_sounds_file(char *sndsfile_name);
extern void  read_soundsfile_sounds();
extern int   read_sound(int slot, char *file);
extern void  unload_sounds_file();
extern void  unload_sound(int slot);


/*__________________PROGRAMMA___________________*/

int main (int argc, char **argv) {
#ifdef SEGFAULT_HANDLER
	signal(SIGSEGV, segfault);
	signal(SIGFPE, segfault);
#endif
	signal(SIGTERM, quit);

	if (argc != 2) {
		printf (GSI_SERVER_VERSION"\n");
		exit (0);
	}

	server_name=argv[0];

#ifndef USE_SOCKETS
	/* Call the plumber to unblock the 'in' pipe! */
	fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
#else
	/* We need to do a setpgrp so that the daemon will no longer be associated
	 * with the user's control terminal. 
	 *
	 * This is done before the fork, so that the child will not be a process
	 * group leader.  Otherwise, if the child were to open a terminal, it
	 * would become associated with that terminal as its control terminal.
	 * It is always best for the parent to do the setpgrp.
	 */
	setpgrp();

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	init_connection();
#endif /* USE_SOCKETS y/n */

	dspdev=DSP_DEV;

	init_mixer();
	snd_server();
	return 0;
}

/*---------------------------------------------------------------------------*
 *	Main server loop. Decodes commands and plays samples.
 *---------------------------------------------------------------------------*/
void snd_server () {
	boolean paused=False;
	int  are_using_dsp=False;

	int  i, j, p;
	int  command;
	byte b;         // pipe/socket data
	int  socket_fd;
	int  client;

	make_sinus(); // for notes, observer rotation etc.

	for(;;) { /* the main loop that we only leave on exit */
		DEB2(fprt(stderr, "main loop\n"));

		add_client(); /* Looks for a new one during play */
		// + ------------------------------------------- +
		for (client=0; client<num_clients; client++) {
			current_client=client;
			current=client_list[client];

			socket_fd=current->socket_fd;

			for (;;) { /* loop to get all sound commands */
				i=read(/*STDIN_FILENO*/ socket_fd,&b,sizeof(b));
				command=b; // => if (command_size==0) command=b;
				DEB1(fprt(stderr, "GSI_SERVER: read status=%d\n", i));
				DEB1(if (i>0) fprt(stderr, "GSI_SERVER: command %d\n", b));

				if (i==0)  {
					/* EOF on the pipe means the parent has closed its end */
					DEB1(fprt(stderr, "%s: EOF from client %d\n", server_name, client));
#ifndef USE_SOCKETS /* i.e. pipe */
					kill(getpid(), SIGTERM);
#else
					remove_client(client);
					remove_dsp_request(client);
					for (i=0; i<dsp_requests; i++)
						if (dsp_request[i]>client) dsp_request[i]--;
					num_clients--;
					client--; break; /* skip this loop and get the next client */
#endif
				}
				if(i!=-1)  {  /* there was something in the pipe */
					DEB2(fprt(stderr, "YES, data for client %d\n", client));
					decode_command(command);
					/* SND_CLOSE_SERVER messes client no up */
					if (command==SND_CLOSE_SERVER) {
						client--; break; /* skip this loop and get the next client */
					}
				}
				else {
					DEB2(fprt(stderr, "No data for client %d\n", client));
					if (errno==EPIPE || errno==EBADF) {
						perror("client dead");
#if 0
					remove_client(client);
					remove_dsp_request(client);
					for (i=0; i<dsp_requests; i++)
						if (dsp_request[i]>client) dsp_request[i]--;
					num_clients--;
					client--; break; /* skip this loop and get the next client */
#endif
					}
					break; /* next client */
				}
			} /* end of loop to get all sound commands */
		} /* loop to handle all clients */

		// -- + -------------------------------------------------------- +
		//send all callback info (only so much that we don't need to wait)
		status_flush();

		are_using_dsp=False;
		for (i=0; i<num_clients; i++) {
			if (client_list[i]->playnum || client_list[i]->channel_playnum) {
				are_using_dsp=True;
				break;
			}
		}
		DEB2(fprt(stderr, "are_using_dsp=%d\n", are_using_dsp));

		if (!(dsp >0 && are_using_dsp)) {
			int  m, max_fd;
			fd_set read_fd;

			DEB1(fprt(stderr, "waiting\n"));
			FD_ZERO(&read_fd);
			FD_SET(ls, &read_fd); /* to catch connection requests (and disconnects) */
			max_fd=ls;
			for (m=0; m<num_clients; m++) {
				socket_fd=client_list[m]->socket_fd;

				FD_SET(socket_fd, &read_fd);
				max_fd=max(max_fd, socket_fd);
			}
			select(max_fd+1, &read_fd, NULL, NULL, NULL);
//			if (FD_ISSET(ls, &read_fd)) {
//				if (add_client()==-1) {
					/* check for disconnect we simply reuse the above code that gets executed when playing samples */
//					continue; /* Hmm, could then remove the addclient here! */
//				}
//			}
			DEB1(fprt(stderr, " ->wait over\n"));
			continue;
		}

// Perhaps for a non blocking write:
// ---------------------------------
//		check available space (probably needed when midi playing is included)
//		int space;
//		ioctl(read_fd, SNDCTL_DSP_GET_OSPACE, &space);
//		if (space <dsp_fragsize) continue;

		if (!are_using_dsp && autorelease_dsp && dsp>-1) {
			close(dsp); dsp=-1;
			DEB1(fprt(stderr, "GSI_SERVER: autorelease: dsp = -1\n"));
		}

#ifdef USE_OSS /* Midi only for OSS at the moment */
		if (playsong) {
			play_midi_events(250000); /* max 0.25 s */
			/* could put this in play_midi_events */
			last_midi_write += (102.4 * (double) dsp_samples)/(double) dsp_speed;
/* current midi=current_time?? */
			if (current_midi>last_midi_write) last_midi_write = current_midi;
		}
		if (_seqbufptr) { SEQ_DUMPBUF(); }
#endif

		// Now play all the active sounds
		if (are_using_dsp) {
			/* ** Mix the sounds into the final buffer */
			memset(premix, 0, sizeof(premix));

			for (current_client=0; current_client<num_clients; current_client++) {
				current=client_list[current_client];
				for(i=0; i<current->playnum; i++) {
					play_fragment(i);
				}
				for(i=0; i<current->channel_playnum; i++) {
					play_fragment(current->channel_used[i]+MAX_DYNAMIC_CHANNELS);
				}
			}

			/* clip the sound size from 32 bit (16 bit + overflow) to 
			 * 16 bit or 8 bit */
			switch(dsp_type) {
			case GSI_8BIT:
				DEB2(fprintf(stderr, "GSI_8BIT device\n"));
				for(i=0;i<dsp_samples *dsp_channels;i++) {
					int pmb = premix[i]>>8 /* 9 */;
					final[i] = (pmb>127) ? 255 : (pmb<-128) ? 0 : (pmb ^ 128);
				}
				break;
			case GSI_8BIT_MULAW:
				DEB2(fprintf(stderr, "GSI_8BIT_MULAW device\n"));
				for(i=0;i<dsp_samples *dsp_channels;i++) {
					int final_tmp=premix[i];
					final[i]= linear16_to_mulaw(final_tmp /*>>1*/);
				}
				break;
			case GSI_16BIT:
				DEB2(fprintf(stderr, "GSI_16BIT device\n"));
				for(i=0;i<dsp_samples *dsp_channels;i++) {
					int final_tmp=premix[i] /*>>1*/;
					/* assuming little endian sound device */
					final[i<<1]= final_tmp &0xff;
					final[(i<<1)+1] = (final_tmp>>8) &0xff;
				}
				break;
			}

			/* check end of files -> if so, stop/repeat it (i.e. repeat's
			 * are only done per fragment, up to a fragsamples-1 samples
			 * silence between; change? YES) */

			/* SHOULD THIS CODE NOT BE IN THE play_fragment()
			   (OR IMMEDIATELY AFTER?) */
			for (client=0; client<num_clients; client++) {
				current_client=client;
				current=client_list[client];

				playnum=current->playnum;

				for(i=playnum-1; i>=0; i--) {
					if (current->position[i]>=current->sound_size[current->playing[i]]) {
						DEB1(fprt(stderr, "GSI_SERVER: finished playing sound %d (queue no %d)\n", current->playing[i], i));
						DEB1(fprt(stderr, "  it was at position %d\n", current->position[i]));
						if (current->sound_flags[i] & SF_REPEAT) {
							current->position[i]=0;
							DEB1(fprt(stderr, "  repeating it"));
						}
						else {
							if  ( i<playnum-1 ) {
								if (i<current->last_sound_stack)
									current->last_sound_stack--;
								else if (i==current->last_sound_stack)
									current->last_sound_stack = -1;
								memmove(current->sound_flags+i, current->sound_flags+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->playing+i, current->playing+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->position+i, current->position+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_x+i, current->sound_x+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dx+i, current->sound_dx+i+1, (playnum-i-1)*sizeof(short));
								memmove(current->sound_z+i, current->sound_z+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dz+i, current->sound_dz+i+1, (playnum-i-1)*sizeof(short));
								memmove(current->sound_y+i, current->sound_y+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dy+i, current->sound_dy+i+1, (playnum-i-1)*sizeof(short));
							}
							playnum--;
							current->playnum--;
						}
					}
				}

				/* Same for channels */
				channel_playnum=current->channel_playnum;
				for(i=channel_playnum-1; i>=0; i--) {
					int  j=MAX_DYNAMIC_CHANNELS+current->channel_used[i];
					if (current->position[j]>=current->sound_size[current->playing[j]]) {
						DEB1(fprt(stderr, "GSI_SERVER: finished playing sound %d (channel no %d)\n", current->playing[j], current->channel_used[i]));
						DEB1(fprt(stderr, "  it was at position %d\n", current->position[j]));
						if (current->sound_flags[j] & SF_REPEAT) {
							current->position[j]=0;
							DEB1(fprt(stderr, "  repeating it"));
						}
						else {
							if  ( i<channel_playnum-1 ) {
								if (i<current->last_sound_channel)
									current->last_sound_channel--;
								else if (i==current->last_sound_channel)
									current->last_sound_channel=-1;
								memmove(current->channel_used+i, current->channel_used+i+1, (channel_playnum-i-1)*sizeof(int));
							}
							channel_playnum--;
							current->channel_playnum--;
							current->playing[j]=-1; /* used as a flag to check if a sound is playin on the channel, should use a separate status_flag[MAX_CHANNELS] byte/int  */
							if (current->do_callback) {
								status_send_byte(GSI_CALLBACK_CHANNEL_END);
								status_send_byte(current->channel_used[i]);
								DEB1(fprintf(stderr, "GSI_SERVER: callback in queue\n"));
							}
						}
					}
				}
			}
		} else {
			/* No sounds to play so fill the buffer with silence */
			/* Should do this too before releasing the DSP ? */
			if (dsp_type==GSI_8BIT) {
				memset(final, 128, dsp_fragsize);
			}
			else if (dsp_type==GSI_8BIT_MULAW) {
				memset(final, 255, dsp_fragsize);
			}
			else { /* 16 bit */
				memset(final, 0, dsp_fragsize);
			}
		}
		write (dsp, final, dsp_fragsize); /* blocking write */
		current_time = gettimeofday_diff();
	}
} // ^ sndserver






/*---------------------------------------------------------------------------*
 *  Sound flag manipulation for 3d
 *---------------------------------------------------------------------------*/
void check_dxyz(int j) {
	DEB1(fprintf(stderr, "check_dxyz for %d\n", j));
	if (current->sound_dx[j] | current->sound_dz[j] | current->sound_dy[j]) {
		current->sound_flags[j] |= SF_RECALC_3D;
		DEB1(fprintf(stderr, "3d!\n"));
	}
	else
		current->sound_flags[j] &= ~SF_RECALC_3D;
}



/*---------------------------------------------------------------------------*
 *  calculate one fragment of a given sound
 *---------------------------------------------------------------------------*/
void play_fragment(int i) {
	int  j;

	int  sound_volume_L, sound_volume_R; // stereo position volume factors

// Volume factors for stereo sounds (is this useful? Perhaps only in case of music)
//int  sound_volume_L_L, sound_volume_L_R;
//int  sound_volume_R_L, sound_volume_R_R;

//	stereo_shift=channels-sound_channels[i];
	char *sample_position;
	int  sample_length;
	int  conversion;
	int  sound_conv;
	int  sound_conv2;
	/* == 3d system */
	int  radius, radius_sq; /* x-z plane distance */
	int  dist, dist_sq; /* the 3d distance */
	int  x, z, y; //position without the 8 bit fraction

	int  f; /* doppler effect */

#define USE_DOPPLER

/* ================= new 3d system ==================== */
//	int old_x, old_y, old_z;

	/* First some updating */
	if (current->sound_flags[i] & (SF_RECALC_3D | SF_RECALC_3D_ONCE | SF_INIT_POSITION) ) {
		int abs_x, abs_y;
		DEB2(fprintf(stderr, "3d update\n"));
// Should I put the time adjust here instead of in the init at SND_SET_X etc?
// First use this:
//      dt=current_time-last_position_time[j];
// Then:
//		last_position_time[j]=current_time;

//old_x=sound_x[i]; old_y=sound_y[i]; old_z=sound_z[i];
//		if (sound_flags[i] & (SF_RECALC_3D | SF_RECALC_3D_ONCE)) {
		DEB2(fprintf(stderr, "(sound_dx[i]*32*dsp_samples)/dsp_speed=%d\n", (current->sound_dx[i]*32*dsp_samples)/dsp_speed));
		current->sound_x[i] += (current->sound_dx[i]*32*dsp_samples)/dsp_speed; /* This corrects the change which is designed for 1/32s to the actual time used per fragment */
		current->sound_z[i] += (current->sound_dz[i]*32*dsp_samples)/dsp_speed;
		current->sound_y[i] += (current->sound_dy[i]*32*dsp_samples)/dsp_speed;
//Could approximate these by sound_x[i] += (sound_dx[i]*g)>>8 etc, where g=(32*dsp_samples<<8)/dsp_speed; (as in frame time after dsp write)
		x=current->sound_x[i]>>8; /* remove the 8 bit fraction */
		z=current->sound_z[i]>>8;
		y=current->sound_y[i]>>8;

		//now  calc r, pan, ...
		radius_sq = x*x+z*z;
		dist_sq = radius_sq  + y*y;
		abs_x=abs(x); abs_y=abs(y);
		radius = max(abs_x, abs_y)+((min(abs_x, abs_y)*5)>>4); /* here I approximated n/3 by (n*5>>4) */
		if (radius>0)
			radius = (radius + radius_sq/radius) >>1;
		else radius=1;
		current->sound_pan[i] = (x<<8)/radius;
		dist= sqrt(dist_sq); /* could use radius with y and the same algo as above */

/*---------------------------------------------------------------------------*
 * 3d: R = x,z-plane radius
 *	        1-(x/R)                                1+(x/R)  
 *  vol_L= -------- (r0/r)^2 *fspeed*vol,  vol_R= -------- (r0/r)^2 *fspeed*vol
 *            2                                      2
 *
 * Pan:
 *         (1-pan/256)                (1+pan/256)
 *  vol_L= ----------- *vol    vol_R= ----------- *vol
 *              2                          2
 *---------------------------------------------------------------------------*/

		current->sound_flags[i] &= ~SF_INIT_POSITION;
	} else { dist=current->old_dist[i]; dist_sq=dist*dist; }
	DEB2(fprintf(stderr, "3d update done\n"));
//	mono & stereo
//		sound_volume_L= ((sound_volume[i]>>1)*r0*r0)/dist_sq;
#define r0 64
	// stereo:
	sound_volume_L= ( ((((256*256 - (current->sound_pan[i]*current->base_width))) *current->sound_volume[i])>>16) *r0*r0)/dist_sq; /* incomplete, f */
	sound_volume_R= ( ((((256*256 + (current->sound_pan[i]*current->base_width))) *current->sound_volume[i])>>16) *r0*r0)/dist_sq; /* incomplete, f */
	//For mono, use (sound_volume_L+sound_volume_R)/2
	current->sound_flags[i] &= ~SF_RECALC_3D_ONCE;
#undef r0
	DEB2(fprintf(stderr," v=%d\n", (((256*256 - (current->sound_pan[i]*current->base_width))) *current->sound_volume[i])));

	DEB2(fprt(stderr, "sound %d, pan=%d, vol_L=%d, vol_R=%d, dist_sq=%d\n", current->playing[i], current->sound_pan[i], sound_volume_L, sound_volume_R, dist_sq));


#ifdef USE_DOPPLER
	// doppler_freq=freq* vg/(vg+(r'-r)/dt)
	if (current->sound_flags[i] & SF_DOPPLER) {
//Not accurate enough:
//		f=(mach1<<8)/(mach1 + (old_dist[i]-dist));

//Therefore:
		int A;
//			A=sound_dx[i]*(sound_dx[i]+old_x*2) + sound_dy[i]*(sound_dy[i]+old_y*2)  + sound_dy[i]*(sound_dy[i]+old_y*2);
//x,y,z have been updated, hence:
		A = current->sound_dx[i]*(((current->sound_x[i]<<1)-current->sound_dx[i])>>8)+current->sound_dy[i]*(((current->sound_y[i]<<1)-current->sound_dy[i])>>8)+current->sound_dz[i]*(((current->sound_z[i]<<1)-current->sound_dz[i])>>8);
//Should correct A with a factor (32*dsp_samples)/dsp_speed as above in the x,y,z update. Also use this for current_time update
		f=(current->mach1<<8)/(current->mach1 - (((32*A)/current->old_dist[i])>>1));

		DEB2(fprintf(stderr, "A=%d, f=%d, (((32*A)/old_dist[i])>>1)=%d\n", A, f, (((32*A)/current->old_dist[i])>>1)));
		DEB2(fprintf(stderr, "mach1=%d\n, old_dist[i]=%d, dist=%d\n", current->mach1, current->old_dist[i], dist));
	}
	else f=256;
	/* we're done with radius, so now set old_dist */
	current->old_dist[i]=dist;

	DEB2(fprintf(stderr, "doppler, f=%d\n", f));
	sound_conv = ( ((current->sound_speed[current->playing[i]]<<8)/dsp_speed)*256) /f;
	sound_conv2= ( ((dsp_speed<<8)/current->sound_speed[current->playing[i]])* f) /256;


#else /* ! USE_DOPPLER */
	int  sound_conv=current->sound_conversion[current->playing[i]];
	int  sound_conv2=(dsp_speed<<8)/current->sound_speed[current->playing[i]];
#endif /* USE_DOPPLER y/n */

	sample_position=current->sound_buffer[current->playing[i]]+current->position[i];

	DEB2(fprt(stderr, "sound %d, sound conv=%d, sound_conv2=%d\n", current->playing[i], sound_conv, sound_conv2));
	DEB2(fprt(stderr, "  -> sample_position %d\n", current->position[i]));
	DEB2(fprt(stderr, "  -> stereo_pos=%d, stereo_speed=%d, observer_x=%d\n", current->sound_x[i], current->sound_dx[i], current->observer_x));
	DEB2(fprt(stderr, "  -> vol L,R=%d, %d\n", sound_volume_L, sound_volume_R));

//	int sample_length=min(dsp_fragsize, (((sound_size[playing[i]]-position[i])*sound_conv2)>>8));
//	int sample_length=min(dsp_samples, (((sound_size[playing[i]]-position[i])*sound_conv2)>>8) * (dsp_bps*dsp_channels) );

	sample_length=min(dsp_samples, (((current->sound_size[current->playing[i]]-current->position[i])*sound_conv2)>>8) / (current->sound_channels[current->playing[i]]* ((current->sound_type[current->playing[i]] +1)/2) ));


	conversion = (current->sound_type[current->playing[i]] - 1)<<2;
	conversion |= (current->sound_channels[current->playing[i]]-1)<<1;
	conversion |= (dsp_channels-1);

	if (dsp_channels==1) sound_volume_L = (sound_volume_L+sound_volume_R)>>1;
	DEB2(fprt(stderr, "  -> vol L=%d\n", sound_volume_L));
//		sound_volume_L=(sound_volume_L+sound_volume_R)>>1;
//should really use two calc's above so we don't calc L and R for mono DSP.

/*
 * switch to 16 bit internal always. This halves the number of routines and
 * is even faster for 8 bit sounds (no >>8 after a volume mul !)
 *
 * Bits          sound type | stereo/mono sound | stereo/mono dsp */

#define S16_M  (((2<<2)|(1<<1)) | 0)
#define  S8_M  (((1<<2)|(1<<1)) | 0)
#define  SU_M  (((0<<2)|(1<<1)) | 0)
#define M16_M  (((2<<2)|(0<<1)) | 0)
#define  M8_M  (((1<<2)|(0<<1)) | 0)
#define  MU_M  (((0<<2)|(0<<1)) | 0)
#define S16_S  (((2<<2)|(1<<1)) | 1)
#define  S8_S  (((1<<2)|(1<<1)) | 1)
#define  SU_S  (((0<<2)|(1<<1)) | 1)
#define M16_S  (((2<<2)|(0<<1)) | 1)
#define  M8_S  (((1<<2)|(0<<1)) | 1)
#define  MU_S  (((0<<2)|(0<<1)) | 1)

	switch(conversion) {
		// Always 16 bit premix, hence, with 8 bit samples do <<8, and in
		// the final mixing, decide if >>8 is needed. Software volumes in
		// the range 0-256 actually remove the <<8 for 8 bit samples.
	case S16_M: /* stereo 16 bit sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "S16_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		for(j=0;j<sample_length;j++) {
			int16 b1, b2;
			b1  = ((int) *(sample_position+(((sound_conv*j)>>6) & ~3)+0))<<8;
			b1 |= (byte) *(sample_position+(((sound_conv*j)>>6) & ~3)+1);
			b2  = ((int) *(sample_position+(((sound_conv*j)>>6) & ~3)+2))<<8;
			b2 |= (byte) *(sample_position+(((sound_conv*j)>>6) & ~3)+3);
			premix[j] += ((b1+b2)*sound_volume_L)>>(1+0+8);
		}
		current->position[i] += ((dsp_samples*sound_conv)>>6) & ~3;
		break;

	case SU_M:  /* stereo 8 bit mulaw sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "SU_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		for(j=0;j<sample_length;j++) {
			int16 b1, b2;
			b1  = mulaw_to_linear16( (byte) *(sample_position+(((sound_conv*j)>>7)&~1)+0) );
			b2  = mulaw_to_linear16( (byte) *(sample_position+(((sound_conv*j)>>7)&~1)+1) );
			premix[j] += (((int) b1+ (int) b2)*sound_volume_L)>>9;
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;

	case S8_M:  /* stereo 8 bit sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "S8_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		for(j=0;j<sample_length;j++) {
			int16 b1, b2;
			b1  = (int16) *(sample_position+(((sound_conv*j)>>7)&~1)+0);
			b2  = (int16) *(sample_position+(((sound_conv*j)>>7)&~1)+1);
			premix[j] += (((int) b1+ (int) b2)*sound_volume_L) >>1;
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;


	case M16_M:  /* mono 16 bit sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "M16_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		for(j=0;j<sample_length;j++) {
			int16 b1;
			b1  = ((int) (*(sample_position+(((sound_conv*j)>>7) & ~1)+0))) <<8;
			b1 |=  (byte) *(sample_position+(((sound_conv*j)>>7) & ~1)+1);
			premix[j] += (((int) b1)*sound_volume_L)>>8;
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;

	case MU_M:  /* mono 8 bit mulaw sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "MU_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		if (sound_conv==256) { //no doppler or speed conversion, faster code
			for(j=0;j<sample_length;j++) {
				premix[j] += ( ((int) mulaw_to_linear16( (int) (byte) (*(sample_position+(j))) )) * sound_volume_L) >>8;
			}
		}
		else {
			for(j=0;j<sample_length;j++) {
				premix[j] += ( ((int) mulaw_to_linear16( (int) (byte) (*(sample_position+((sound_conv*j)>>8))) )) * sound_volume_L) >>8;
			}
		}
		current->position[i] += ((dsp_samples*sound_conv)>>8);
		break;

	case M8_M:  /* mono 8 bit sound to mono 16 bit buffer */
		DEB2(fprt(stderr, "M8_M, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		if (sound_conv==256) { //no doppler or speed conversion, faster code
			for(j=0;j<sample_length;j++) {
				premix[j] +=(*(sample_position+(j))) * sound_volume_L;
			}
		}
		else {
			for(j=0;j<sample_length;j++) {
				premix[j] +=(*(sample_position+((sound_conv*j)>>8))) * sound_volume_L;
			}
		}
		current->position[i] += ((dsp_samples*sound_conv)>>8);
		break;

#if 0 /* reintegrate this in the new playing code */
//oversample code v2
//Standard interpolation !!this code gives good results!!
			{
				int delta=((sound_conv*j)&255);
				int premix_plus = ( (256-delta) * ((int)*(sample_position+((sound_conv*j)>>8))) + delta* (int)*(sample_position+((sound_conv*j)>>8)+1) )>>8;
				premix[j] +=premix_plus;

//oversample v3, unfinished
			{
				int new_item=*(sample_position+((sound_conv*j)>>8));
				sample_factor=(float) sound_conv /256.0;

				prev_item = new_item;
				new_item = oversample;
				oversample += oversample_inc;
				if (sample_position+sample_factor > last_offset + 1) {
					oversample_inc = (prev_item-oversample)*sample_factor;
				}
				premix[i] += ((new_item*volume[playing[i]])>>8);
			}
#endif /* to be reintegrated */


	case S16_S: /* stereo 16 bit sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "S16_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		for(j=0;j<sample_length*2;j+=2) {
			int b1, b2;
			b1  = ((int) *(sample_position+(((sound_conv*j)>>7) & ~3)+0))<<8;
			b1 |= (byte) *(sample_position+(((sound_conv*j)>>7) & ~3)+1);
			b2  = ((int) *(sample_position+(((sound_conv*j)>>7) & ~3)+2))<<8;
			b2 |= (byte) *(sample_position+(((sound_conv*j)>>7) & ~3)+3);
			premix[j]   += (b1*sound_volume_L)>>(0+8);
			premix[j+1] += (b2*sound_volume_R)>>(0+8);
		}
		current->position[i] += ((dsp_samples*sound_conv)>>6) & ~3;
		break;

	case SU_S:  /* stereo 8 bit mulaw sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "S8_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

//if (f==256) { faster code
		for(j=0;j<sample_length*2;j+=2) {
			int16 b1, b2;
			b1  = mulaw_to_linear16( (int) (byte) *(sample_position+(((sound_conv*j)>>8)&~1)+0) );
			b2  = mulaw_to_linear16( (int) (byte) *(sample_position+(((sound_conv*j)>>8)&~1)+1) );
			premix[j]  += (b1*sound_volume_L)>>8;
			premix[j+1]+= (b2*sound_volume_R)>>8;
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;

	case S8_S:  /* stereo 8 bit sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "S8_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

//if (f==256) { faster code
		for(j=0;j<sample_length*2;j+=2) {
			int16 b1, b2;
			b1  = (int16) *(sample_position+(((sound_conv*j)>>8)&~1)+0);
			b2  = (int16) *(sample_position+(((sound_conv*j)>>8)&~1)+1);
			premix[j]  += b1*sound_volume_L;
			premix[j+1]+= b2*sound_volume_R;
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;

	case M16_S:  /* mono 16 bit sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "M16_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		if (sound_conv==256) {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = ( (int) (*(sample_position+((j) & ~1)+0)) ) <<8;
				b1 |=   (byte) *(sample_position+((j) & ~1)+1);
				premix[j]   += (((int) b1) *sound_volume_L)>>8;
				premix[j+1] += (((int) b1) *sound_volume_R)>>8;
			}
		}
		else {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = ((int16) (*(sample_position+(((sound_conv*j)>>8) & ~1)+0))) <<8;
				b1 |=  (byte) *(sample_position+(((sound_conv*j)>>8) & ~1)+1);
				premix[j]   += (((int) b1) *sound_volume_L)>>8;
				premix[j+1] += (((int) b1) *sound_volume_R)>>8;
			}
		}
		current->position[i] += ((dsp_samples*sound_conv)>>7) & ~1;
		break;

	case MU_S:  /* mono 8 bit mulaw sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "MU_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		if (sound_conv==256) {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = mulaw_to_linear16( (int) (byte) (*(sample_position+(j>>1) )) );
				premix[j]   += (((int) b1) * sound_volume_L)>>8;
				premix[j+1] += (((int) b1) * sound_volume_R)>>8;
			}
		}
		else {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = mulaw_to_linear16( (int) (byte) (*(sample_position+(((sound_conv*j)>>9) ))) );
				premix[j]   += (((int) b1) * sound_volume_L)>>8;
				premix[j+1] += (((int) b1) * sound_volume_R)>>8;
			}
		}
		current->position[i] += ((dsp_samples*sound_conv)>>8);
		break;

	case M8_S:  /* mono 8 bit sound to stereo 16 bit buffer */
		DEB2(fprt(stderr, "M8_S, sound %d (snd file %d), sample length=%d\n", i, current->playing[i], sample_length));

		if (sound_conv==256) {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = ((int) (*(sample_position+(j>>1) )));
				premix[j]   += ((int) b1) * sound_volume_L;
				premix[j+1] += ((int) b1) * sound_volume_R;
			}
		}
		else {
			for(j=0;j<sample_length*2;j+=2) {
				int16 b1;
				b1  = ((int) (*(sample_position+(((sound_conv*j)>>9) ))) );
				premix[j]   += ((int) b1) * sound_volume_L;
				premix[j+1] += ((int) b1) * sound_volume_R;
			}
		}
		current->position[i] += ((dsp_samples*sound_conv)>>8);
		break;
	}
}


/*---------------------------------------------------------------------------*
 *  Remove a grab dsp request from the given client (quit or SND_RELEASE_DSP)
 *---------------------------------------------------------------------------*/
void remove_dsp_request(int client) {
	int  i;

	DEB1(fprt(stderr, "removing dsp request for %d\n", client));
	if (dsp_requests>0) {
		if (dsp_request[0]==client) {
			/* same code as when client dsp_request[0] gives up the DSP: */
			if (dsp_requests>1) { /* another client wants the DSP */
				struct client_data  *cd = client_list[dsp_request[1]];

				/* Should use delayed dsp setup when we are playing sounds */
				dsp=setup_dsp(dspdev, cd->new_dsp_speed, cd->new_dsp_log2_frag, cd->new_dsp_channels, cd->new_dsp_type);

				autorelease_dsp=cd->autorelease_dsp;
				autoset_dsp=cd->autoset_dsp;
			}
			else { /* No other clients so free the DSP */
				autorelease_dsp=False;
				autoset_dsp=False; /* Correct? or with SET_DSP_SPEED etc? */
				close(dsp); dsp=-1;
			}
		}
		for (i=0; i<dsp_requests; i++) {
			if (dsp_request[i]==client) {
				if (dsp_requests-i-1>0)
					memmove(&dsp_request[i], &dsp_request[i+1], dsp_requests-i-1);
				dsp_requests--;
				break;
			}
		}
	}
}


/*---------------------------------------------------------------------------*
 *  Decode commands we got from pipe/socket
 *---------------------------------------------------------------------------*/
void decode_command (int command) {
	int  p; /* used for channel number etc */
	int  i;
	int  j;

	// + ------------------------------------------- +
	//Command decoding:
//ASYNC:		command_data[sound_size-1]=get_byte(); /* This is WRONG, fix */
//ASYNC:		DEB1(fprt(stderr, "GSI_SERVER: command=%d, data=%d\n", command, command_data));
	switch(command) {
	case SND_OPEN_SERVER:
		break;
	case SND_CLOSE_SERVER:
		remove_client(current_client);
		remove_dsp_request(current_client);
		for (i=0; i<dsp_requests; i++)
			if (dsp_request[i]>current_client) dsp_request[i]--;
		num_clients--;
		break;
	case SND_GRAB_DSP:
		DEB1(fprt(stderr, "GSI_SERVER: SND_GRAB_DSP\n"));
		if (current->dsp_request==SND_RELEASE_DSP) { /* => new request */
			dsp_request[dsp_requests]=current_client;
			dsp_requests++;
			DEB1(fprt(stderr, "  GRAB_DSP: dsp_requests=%d, current_client=%d\n", dsp_requests, current_client));
		}
		current->dsp_request=SND_GRAB_DSP;
		current->autorelease_dsp=False;
		current->autoset_dsp=False;

		if (dsp_request[0]==current_client) {
			dsp=setup_dsp(dspdev, current->new_dsp_speed, current->new_dsp_log2_frag, current->new_dsp_channels, current->new_dsp_type);
			autorelease_dsp=False;
			autoset_dsp=False; /* Correct? or with SET_DSP_SPEED etc? */
		}
		break;
	case SND_RELEASE_DSP:
		DEB1(fprt(stderr, "GSI_SERVER: SND_RELEASE_DSP\n"));
		current->dsp_request=SND_RELEASE_DSP;
		current->autorelease_dsp=False;
		remove_dsp_request(current_client);
		break;
	case SND_AUTO_DSP:
		DEB1(fprt(stderr, "GSI_SERVER: SND_AUTO_DSP\n"));
		current->dsp_request=SND_AUTO_DSP;
		current->autorelease_dsp=True;
		if (dsp_request[0]==current_client) autorelease_dsp=True;
		break;
	case SND_USE_BEST_DSP_SETTINGS: //rename: SND_AUTO_SET_DSP / SND_AUTO_CONFIG_DSP ?
		DEB1(fprt(stderr, "GSI_SERVER: SND_USE_BEST_DSP_SETTINGS\n"));
		current->autoset_dsp=True;
		if (dsp_request[0]==current_client) autoset_dsp=True;
		break;
	case SND_SET_DSP_SPEED:
//command_data[sound_size++ -1]=get_byte();
//ASYNC:			if (command_size<3) continue;
//ASYNC:			new_dsp_speed=command_data[0] | (command_data[1]<<8);
//ASYNC:			command_size=0;
		DEB1(fprt(stderr, "GSI_SERVER: SND_SET_DSP_SPEED\n"));
		current->new_dsp_speed = (int) (uint16) get_int16();
		DEB1(fprt(stderr, "           set to %d\n", current->new_dsp_speed));

		break;
	case SND_SET_DSP_CHANNELS:
		current->new_dsp_channels = (int) get_byte();
//ASYNC:			if (command_size<1) continue;
//ASYNC:			new_dsp_channels=command_data[0];
//ASYNC:			command_size=0;
		break;
	case SND_SET_DSP_BYTES: //should be SND_SET_DSP_TYPE
		current->new_dsp_type = (int) get_byte();
//ASYNC:			if (command_size<1) continue;
//ASYNC:			new_dsp_type=command_data[0];
//ASYNC:			command_size=0;
		break;
	case SND_SET_DSP_FRAGSIZE:
		current->new_dsp_log2_frag = get_byte();
//ASYNC:			if (command_size<1) continue;
//ASYNC:			new_dsp_log2_frag = command_data[0];
//ASYNC:			command_size=0;
		break;

	case SND_SET_MAX_AUTO_CHANNELS:
//ASYNC:			if (command_size<1) continue;
		current->max_stack=min(get_byte(), MAX_DYNAMIC_CHANNELS);
		break;

	case SND_LOAD_SOUND_LIST:
		DEB1(fprt(stderr, "GSI_SERVER: SND_LOAD_SOUND_LIST\n"));
		{
			char *name=get_string();
			init_sounds_file(name);
			free(name);
			break;
		}
	case SND_FREE_SOUND_LIST:
	{
		unload_sounds_file();
		break;
	}
	case SND_LOAD_LISTED:
	{
		read_soundsfile_sounds();
		break;
	}
	case SND_FREE_LISTED:
	{
		unload_sounds_file(); //should not release the list file.
		break;
	}
	case SND_LOAD_SOUND:
	{
		int  slot=get_byte();
		if (current->sound_filename[slot]!=NULL)
			read_sound(slot, current->sound_filename[slot]);
	}
	break;
	case SND_LOAD_SOUND_FILE:
		DEB1(fprt(stderr, "GSI_SERVER: SND_LOAD_SOUND_FILE\n"));
		{
			//Always SYNC mode here
			int  slot=get_byte();
			char *name=get_string();
			if (slot < /*max_slot*/ MAX_SOUNDS) {
				char *name2=malloc(strlen(name)+strlen(current->sounds_dir)+3);
				strcpy(name2, current->sounds_dir); strcat(name2, "/");
				if (name[0]=='/' || strlen(current->sounds_dir)==0)
					strcpy(name2, name);
				else
					strcat(name2, name);
				current->num_sounds=max(current->num_sounds, slot+1);
				read_sound(slot, name2);
				free(name2);
			}
			free(name);
			break;
		}
	case SND_LOAD_SOUND_DATA:
		DEB1(fprt(stderr, "GSI_SERVER: SND_LOAD_SOUND_DATA\n"));
		{
			//Always SYNC mode here
			int  l;
			int  slot=get_byte();
			int  sample_speed=get_int32();
			int  sample_type=get_byte();
			int  sample_channels=get_byte();
			int  sample_len=get_int32();
			if (slot < /*max_slot*/ MAX_SOUNDS) {
				unload_sound(slot);

				current->sound_buffer[slot]=(byte *) malloc(sample_len * sample_channels * ((sample_type+1)>>1));
				get_block(current->sound_buffer[slot], sample_len);

				current->num_sounds=max(current->num_sounds, slot+1);
				init_sample(slot, sample_speed, sample_type, sample_channels, sample_len);
			}
			else for (l=0; l<sample_len; l++) get_byte(); //ignore all sent
			break;
		}

	case SND_SET_SLOT_INFO:
		DEB1(fprt(stderr, "GSI_SERVER: SND_SET_SLOT_INFO\n"));
		{
			int  slot=get_byte();
			int  sample_speed=get_int32();
			int  sample_type=get_byte();
			int  sample_channels=get_byte();

			init_sample(slot, sample_speed, sample_type, sample_channels, current->sound_size[slot]);
		}
	case SND_GET_SLOT_INFO:
		DEB1(fprt(stderr, "GSI_SERVER: SND_GET_SLOT_INFO not implemented\n"));
		break;

	case SND_LOAD_SONG_FILE:
		DEB1(fprt(stderr, "GSI_SERVER: SND_LOAD_SONG_FILE\n"));
		{
			//Always SYNC mode here
			char *name=get_string();
			if (midifilebuf !=NULL) free(midifilebuf);
			if (load_song(name)==0) fprintf(stderr, "Bad midi file %s\n", name);
			free(name);
			playsong=0; playsong_ct=0;
			break;
		}
	case SND_LOAD_SONG_DATA:
		DEB1(fprt(stderr, "GSI_SERVER: SND_LOAD_SONG_DATA\n"));
		{
			//Always SYNC mode here
			int  l;
			int  sample_len=get_int32();
			if (midifilebuf !=NULL) free(midifilebuf);
//					unload_song();
			midifilebuf=(byte *) malloc(sample_len);
			if (midifilebuf!=NULL)
				get_block(midifilebuf, sample_len);
			else
				for (l=0; l<sample_len; l++) get_byte(); //ignore all sent
//					init_song(); /* check for a valid MIDI etc */
			midi_to_tracks(midifilebuf, sample_len);
			playsong=0; playsong_ct=0;
			break;
		}

	case SND_FREE_SOUND:
		DEB1(fprt(stderr, "GSI_SERVER: SND_FREE_SOUND\n"));
//ASYNC:			if (command_size<1) continue;
		{
//ASYNC:				slot = command_data[0];
//ASYNC:				command_size=0;
			int  slot=get_byte();
			unload_sound(slot);
		}
		break;
	case SND_SET_SOUND_DIR:
		DEB1(fprt(stderr, "GSI_SERVER: SET_SOUND_DIR\n"));
		{ //SYNC
			char *name=get_string();
			DEB1(fprt(stderr, "  sounds_dir=%s\n", name));
			strncpy(current->sounds_dir, name, 255);
			current->sounds_dir[255]=0;
			free(name);
			break;
		}
	case SND_CALLBACK_ON:
		DEB1(fprt(stderr, "GSI_SERVER: callback on\n"));
		current->do_callback=True; break;
	case SND_CALLBACK_OFF:
		DEB1(fprt(stderr, "GSI_SERVER: callback off\n"));
		current->do_callback=False; break;

	case SND_SET_SOUND_SPEED:
		DEB1(fprt(stderr, "GSI_SERVER: SND_SET_SOUND_SPEED (mach1)\n"));
		current->mach1 = get_int16();
		break;
	case SND_SET_BASE_WIDTH:
		DEB1(fprt(stderr, "GSI_SERVER: set base_width\n"));
		current->base_width = get_int16();
		break;
	case SND_SET_MIXER:
		DEB1(fprt(stderr, "GSI_SERVER: SND_SET_MIXER\n"));
		{
			int dev, vol;
			dev =get_byte();
			vol =get_byte();
			set_mixer(dev, vol);
		}
		break;
	case SND_GET_MIXER:
		DEB1(fprt(stderr, "GSI_SERVER: SND_GET_MIXER\n"));
		{
			int dev, vol;
			dev =get_byte();
			vol =read_mixer(dev);
//					send_status(SND_STATUS_MIXER, dev, vol>0? vol-1: 0);
			status_send_byte(GSI_STATUS_MIXER);
			status_send_byte(dev);
			status_send_byte(vol>0? vol-1: 0);
		}
		break;

	case SND_PAUSE_ALL:
		DEB1(fprt(stderr, "GSI_SERVER: SND_PAUSE_ALL\n"));
		current->paused=True;
		break;
	case SND_CONTINUE_ALL:
		DEB1(fprt(stderr, "GSI_SERVER: SND_CONTINUE_ALL\n"));
		current->paused=False;
		break;


	default:
//					if (dsp ==-1 || paused) continue;
		DEB1(fprt(stderr, "decoding:  command is in 0-191\n"));

		if (command == SND_KILLALL) {
			int  pc;
			DEB1(fprt(stderr, "SND_KILLALL\n"));
			current->playnum=0;
			for (pc=0; pc<current->channel_playnum; pc++) 
				current->playing[current->channel_used[pc]+MAX_DYNAMIC_CHANNELS]=-1;
			current->channel_playnum=0;
			if (current->do_callback) status_send_byte(GSI_CALLBACK_CHANNEL_ENDALL);
//			continue;
			return;
		}
		// + ------------------------------------------- +
		if (command <32) { //playing commands
			// Get the channel/sound number
//ASYNC:	p=command_data[0]; /* sound to be played */
			//get <ch> if command >=16
			int  play_channel;
			p=get_byte(); // sound slot
			play_channel=p; // it may be the channel though..
			if (command >=16 && command <16+12) {
				p=get_byte(); // this is the sound slot
				DEB1(fprt(stderr, "GSI_SERVER: channel stuff: play_channel=%d, slot=%d\n", play_channel, p));
			}

			// + ------------------------------------------- +
			if (command<16) { /* Stack channel stuff */
DEB1(fprt(stderr, "GSI_SERVER: commands 0-16\n"));
				playnum=current->playnum;
				if( command == SND_KILL )  {
					DEB1(fprt(stderr, "GSI_SERVER: terminating sound %d\n", (int) p));

					for(i=playnum-1; i>=0; i--) {
						if ( current->playing[i]==p)  {
							if (current->last_sound_stack==i) current->last_sound_stack=-1;
							if  ( i<playnum-1 ) {
								//Hmm, with interleaving I could replace all this with just one memmove! perhaps later (is it worth the bother? must do at least shifts to access a var, muls otherwise)
								memmove(/*repeat*/current->sound_flags+i, /*repeat*/ current->sound_flags+i+1,	(playnum-i-1)*sizeof(int));
								memmove(current->playing+i, current->playing+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->position+i, current->position+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_x+i, current->sound_x+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dx+i, current->sound_dx+i+1, (playnum-i-1)*sizeof(short));
								memmove(current->sound_z+i, current->sound_z+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dz+i, current->sound_dz+i+1, (playnum-i-1)*sizeof(short));
								memmove(current->sound_y+i, current->sound_y+i+1, (playnum-i-1)*sizeof(int));
								memmove(current->sound_dy+i, current->sound_dy+i+1, (playnum-i-1)*sizeof(short));
							}
							playnum--;
							current->playnum--;
							break;
						}
					}
					DEB1(if (i<0) fprintf(stderr, "  not found!\n"));
				} else { //sound play/overwrite/loop commands
					if ( playnum< current->max_stack/*MAX_CHANNELS*/ )  {
						//Load sound if necessary (and if possible)
						if ( current->sound_size[(int)p]==0 ) read_sound(p, current->sound_filename[p]);
						if ( current->sound_size[(int)p]>0 )  {

#if 0 /* Should go into the main server loop ? */
							/* autoset dsp after silence? */
							if (!are_using_dsp && autoset_dsp) {
								dsp=setup_dsp(dspdev, current->sound_speed[p], 0, current->sound_channels[p], current->sound_type[p] | 2); /* MULAW -> 16 BIT (OSS on my GUS PNP doesn't handle mulaw output correctly) */
								set_sound_conversion();
								are_using_dsp=True;
							}
#endif

							j=playnum;
//or like this: if ((command & ~2) == SND_OVERWRITE) (more efficient?)
							if (command == SND_OVERWRITE || command == SND_OVERWRITE_REPEAT) {
								for (i=playnum-1; i>=0; i--)
									if (current->playing[i]==p) { j=i; playnum--; current->playnum--; break; /*ugly*/ }
							}
//									repeat   [j] = (command == SND_REPEAT);
							current->sound_flags[j] = (command == SND_REPEAT || command == SND_OVERWRITE_REPEAT ? SF_REPEAT : 0);
							current->position [j] = 0;
							current->playing  [j] = p;
							current->playnum++;
							playnum++;

							//Now set defaults
							current->sound_volume[j] = 256;
							current->sound_x  [j] = 0;
							current->sound_dx [j] = 0;
							current->sound_z  [j] = 64<<8;
							current->sound_dz [j] = 0;
							current->sound_echo_dt[j]   = 0;
							current->sound_echo_fvol[j] = 255;
							current->old_dist[j] = 64;
							current->last_sound_stack=j;
						}
						DEB1(fprt(stderr, "GSI_SERVER: sound %d added to play-queue at %d\n", p, playnum-1));
					}
				}
			} /* ^ commands 0-15 */
			// + ------------------------------------------- +
			else { /* commands 16-31 (STATIC CHANNELS) */
DEB1(fprt(stderr, "GSI_SERVER: commands 16-31\n"));
				channel_playnum=current->channel_playnum;
				if( command == SND_CHANNEL_KILL )  {
					DEB1(fprt(stderr, "GSI_SERVER: terminating sound on channel %d\n", (int) play_channel));

					for(i=channel_playnum-1; i>=0; i--) {
						if ( current->channel_used[i]==play_channel)  {
							DEB1(fprintf(stderr, "GSI_SERVER: found sound, listno=%d\n", i));
							if (current->last_sound_stack==i) current->last_sound_stack=-1;
							if  ( i<channel_playnum-1 ) {
								memmove(current->channel_used+i, current->channel_used+i+1, (channel_playnum-i-1)*sizeof(int));
							}
							current->playing[play_channel+MAX_DYNAMIC_CHANNELS]=-1;
							if (current->do_callback) {
								status_send_byte(GSI_CALLBACK_CHANNEL_END);
								status_send_byte(play_channel);
								DEB1(fprintf(stderr, "callback in queue\n"));
							}
							channel_playnum--;
							current->channel_playnum--;
							break;
						}
					}
					DEB1(if (i<0) fprintf(stderr, "  not found !\n"));
//					continue;
					return;
				} else { // sound play/overwrite/loop commands
					if ( play_channel< MAX_STATIC_CHANNELS )  {
						// Load sound if necessary (and if possible)
						if ( current->sound_size[(int)p]==0 ) read_sound(p, current->sound_filename[p]);
						if ( current->sound_size[(int)p]>0 )  {

#if 1 /* Should go into the main server loop ? */
							/* autoset dsp after silence? */
							if ((current->playnum|current->channel_playnum)==0 && autoset_dsp) {
								dsp=setup_dsp(dspdev, current->sound_speed[p], 0, current->sound_channels[p], current->sound_type[p] | 2); /* MULAW -> 16 BIT */
								set_sound_conversion();
							}
#endif
							j=MAX_DYNAMIC_CHANNELS+play_channel;
							if (current->playing[play_channel+MAX_DYNAMIC_CHANNELS]>=0) {
								// already a sound playing, overwrite?
								if (!(command == SND_OVERWRITE || command == SND_OVERWRITE_REPEAT)) {
//									continue;
									return;
								}
							}
							else { //new channel
								current->channel_used[channel_playnum]=play_channel;
								channel_playnum++;
								current->channel_playnum++;
DEB1(fprt(stderr, "new sound on channel\n"));
								if (current->do_callback) {
									status_send_byte(GSI_CALLBACK_CHANNEL_STARTED);
									status_send_byte(play_channel);
								}
							}
//									repeat   [j] = (command == SND_CHANNEL_REPEAT);
							current->sound_flags[j] = (command == SND_CHANNEL_REPEAT || command == SND_CHANNEL_OVERWRITE_REPEAT ? SF_REPEAT : 0); 
							current->position [j] = 0;
							current->playing  [j] = p;
							//Now set defaults
							current->sound_volume[j] = 256;
							current->sound_x  [j] = 0;
							current->sound_dx [j] = 0;
							current->sound_z  [j] = 64<<8;
							current->sound_dz [j] = 0;
							current->sound_echo_dt[j]   = 0;
							current->sound_echo_fvol[j] = 255;
							current->old_dist[j] = 64;
							current->last_sound_channel=play_channel;
						}
						DEB1(fprt(stderr, "GSI_SERVER: sound %d started on channel %d (channel_used number %d)\n", p, play_channel, channel_playnum));
					}
				}
			} /* ^ commands 16-31 */
		} /* ^ commands 0-31 */

		// + ------------------------------------------- +
		/* Now handle all other commands: sound modifications,
		 * stereo position, distance, doppler etc. */
		else if (command <64) {
			DEB1(fprt(stderr, "  commands 32-63\n"));
			/* commands 32-63 */
			switch (command) {
			case SND_PLAY_SONG:
				DEB1(fprt(stderr, "SND_PLAY_SONG\n"));
				{
					int song_loop=get_byte();
					DEB1(fprt(stderr, "play song %d times\n", song_loop));
//							play_midi();
//play_song("/home/wouter/projecten/gsi/test/music/InstantX.mid");
					reset_song(); //sets the file pointer etc
					play_midi_init();
					playsong=1; playsong_ct=song_loop;
				}
				break;
			case SND_STOP_SONG:
				DEB1(fprt(stderr, "SND_STOP_SONG\n"));
				playsong=0;
//				Must add a 'midi_stop_voices'
#ifdef USE_OSS
				seq_pause();
#endif
				break;
			case SND_CONTINUE_SONG:
				DEB1(fprt(stderr, "SND_CONTINUE_SONG\n"));
				if (playsong_ct) playsong=1;
//				Must add a 'midi_restore_voices();'
#ifdef USE_OSS
				seq_continue();
#endif
				break;
				//all the observer decoding must also go here
			}
		}
		else { /* commands 64-191 */
			int  command2=command;
			int  data2=current->command_data[0];
			command2 &=~(SND_CURRENT | SND_CHANNEL);
//NB. This is not correct for commands 32-63 which have no CURRENT! FIX.
			if (command & SND_CURRENT) {
				DEB1(fprt(stderr, "GSI_SERVER: last stack=%d, last_channel=%d\n", current->last_sound_stack, current->last_sound_channel));
				if (command & SND_CHANNEL)
					j=current->last_sound_channel+MAX_DYNAMIC_CHANNELS;
				else
					j=current->last_sound_stack;
			}
			else {
				p=get_byte();
				if (command & SND_CHANNEL) {
					j=p+MAX_DYNAMIC_CHANNELS;
					current->last_sound_channel=p;
				}
				else { //find the stack entry that we want to change
					int pc;
					for (pc=playnum; pc>=0; pc--) {
						if (current->playing[pc]==p) {
							j=pc;
							break;
						}
					}
					if (pc<0) /*continue*/ return; //slot p is not being played on any stack channel
					current->last_sound_stack=p;
				}
			}


			DEB1(fprt(stderr, "GSI_SERVER: command=%d, command2=%d, data2=%d, j=%d\n", command, command2, data2, j));
//ASYNC:			sound_flags[j]= /* 1<< */ command2;
//To access STATIC channels: j =j+ MAX_DYNAMICS_CHANNELS;
			switch (command2) {
			case SND_SET_X:
				DEB1(fprt(stderr, "SND_SET_X\n"));
//ASYNC:					sound_x[j] = data2;
//ASYNC:					command_size=0;
				current->sound_x[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE | SF_INIT_POSITION;
				current->last_position_time[j]=current_time;
				break;
			case SND_SET_Z:
				DEB1(fprt(stderr, "SND_SET_Z\n"));
//ASYNC:					sound_z[j] = data2;
//							sound_z[j] = get_int16();
				current->sound_z[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE | SF_INIT_POSITION;
				current->last_position_time[j]=current_time;
				break;
			case SND_SET_Y:
				DEB1(fprt(stderr, "SND_SET_Y\n"));
//ASYNC:					sound_y[j] = data2;
//							sound_y[j] = get_int16();
				current->sound_y[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE | SF_INIT_POSITION;
				current->last_position_time[j]=current_time;
				break;
			case SND_ADJUST_X:
				DEB1(fprt(stderr, "SND_ADJUST_X\n"));
//ASYNC:					sound_x[j] = data2;
//ASYNC:					command_size=0;
				current->sound_x[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE;
				current->last_position_time[j]=current_time;
				break;
			case SND_ADJUST_Z:
				DEB1(fprt(stderr, "SND_ADJUST_Z\n"));
//ASYNC:					sound_z[j] = data2;
//							sound_z[j] = get_int16();
				current->sound_z[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE;
				current->last_position_time[j]=current_time;
				break;
			case SND_ADJUST_Y:
				DEB1(fprt(stderr, "SND_ADJUST_Y\n"));
//ASYNC:					sound_y[j] = data2;
//							sound_y[j] = get_int16();
				current->sound_y[j] = ((int) get_int16()) <<8;
				current->sound_flags[j] |= SF_RECALC_3D_ONCE;
				current->last_position_time[j]=current_time;
				break;


			case SND_SET_DX:
				DEB1(fprt(stderr, "SND_DX\n"));
//ASYNC:					sound_dx[j] = data2;
//							sound_dx[j] = (int) (int8) get_byte();
				current->sound_dx[j] = (int) get_int16();
				check_dxyz(j); /* sets SF_RECALC_3D if needed */
				break;
			case SND_SET_DZ:
				DEB1(fprt(stderr, "SND_DZ\n"));
//ASYNC:					sound_dz[j] = data2;
//							sound_dz[j] = (int) (int8) get_byte();
				current->sound_dz[j] = (int) get_int16();
				check_dxyz(j);
				break;
			case SND_SET_DY:
				DEB1(fprt(stderr, "SND_DY\n"));
//ASYNC:					sound_dy[j] = data2;
//							sound_dy[j] = (int) (int8) get_byte();
				current->sound_dy[j] = (int) get_int16();
				check_dxyz(j);
				break;
			case SND_SET_ANGLE:
				DEB1(fprt(stderr, "SND_SET_ANGLE\n"));
//ASYNC:					angle = data2;
				{
					int angle = get_byte();
					int x=current->sound_x[j]>>8;
					int z=current->sound_z[j]>>8;
					int radius;
					int cos_angle, sin_angle;

					cos_angle=sinus_tabel[((64-angle)&255)<<2];
					sin_angle=sinus_tabel[angle<<2];
					if (x==0) radius=z; else radius=sqrt(x*x+z*z); //no slow sqrt when the sound is in start pos.
//use 32bit integer sqrt routine.
					current->old_dist[j]=radius;
					current->sound_x[j] = (radius*cos_angle)>>8;
					current->sound_z[j] = (radius*sin_angle)>>8;
					current->sound_flags[j] |= SF_RECALC_3D_ONCE;
					current->last_position_time[j] = current_time;
					break;
				}

			case SND_SET_ANGLE_R:
				DEB1(fprt(stderr, "SND_SET_ANGLE_R\n"));
//ASYNC:					angle = data2;
				{
					int angle = get_byte();
					int radius= get_int16();
					int cos_angle, sin_angle;

					cos_angle=sinus_tabel[((64-angle)&255)<<2];
					sin_angle=sinus_tabel[angle<<2];
					current->sound_x[j] = (radius*cos_angle)>>8;
					current->sound_z[j] = (radius*sin_angle)>>8;
					current->old_dist[j] = radius;
					current->sound_flags[j] |= SF_RECALC_3D_ONCE;
					current->last_position_time[j] = current_time;
					break;
				}

#if 0
			case SND_SET_ANGLE_DEFAULT_R:
				//sets the angle and standard radius (i.e. no sqrt needed)
				DEB1(fprt(stderr, "SND_SET_ANGLE_DEFAULT_R\n"));
//ASYNC:					angle = data2;
				{
					int angle = get_byte();
					int cos_angle, sin_angle;
					cos_angle=sinus_tabel[((64-angle)&255)<<2];
					sin_angle=sinus_tabel[angle<<2];
					current->sound_x[j] = (64*cos_angle)>>8;
					current->sound_z[j] = (64*sin_angle)>>8;
					current->last_position_time[j]=current_time;
					break;
				}
#endif
			case SND_SET_PAN:
				DEB1(fprt(stderr, "SND_SET_PAN\n"));
				{
					current->sound_pan[j]= get_int16 ();
					break;
				}


			case SND_SET_VOLUME:
				DEB1(fprt(stderr, "SND_SET_VOLUME\n"));
				{
					current->sound_volume[j]= get_byte ();
					if (current->sound_volume[j]>0) current->sound_volume[j]++;
					break;
				}

			case SND_ECHO_ON:
				DEB1(fprt(stderr, "SND_ECHO\n"));
//ASYNC:					if (command_size<3) continue;
//ASYNC:					sound_echo_dt[j] = data2;
//ASYNC:					sound_echo_dt[j] = command_data[1];
				current->sound_echo_dt[j] = (int) (uint) get_int16();
				current->sound_echo_fvol[j] = get_byte();
				break;
			case SND_ECHO_OFF:
				DEB1(fprt(stderr, "SND_ECHO_OFF\n"));
//ASYNC:					sound_echo_dt[j] = 0;
				current->sound_echo_dt[j] = 0;
				break;
			case SND_DOPPLER_ON:
				DEB1(fprt(stderr, "SND_DOPPLER_ON\n"));
//ASYNC:					sound_doppler[j] = data2;
				// f'=f *Vg/(Vg+Vobj). the same factor is used for the volume change (number of J/(m^2 s) increases/decreases with the frequency)
				current->sound_flags[j] |= SF_DOPPLER;
				break;
			case SND_DOPPLER_OFF:
				DEB1(fprt(stderr, "SND_DOPPLER_OFF\n"));
//for all channels or just one?
				current->sound_flags[j] &= ~SF_DOPPLER;
				break;

			case SND_SET_O_X:
				DEB1(fprt(stderr, "SND_O_X\n"));
//ASYNC:					observer_x = data2;
//							observer_x = get_int16();
				current->observer_x = ((int) get_int16()) <<8;
				break;
			case SND_SET_O_Z:
				DEB1(fprt(stderr, "SND_O_Z\n"));
//ASYNC:					observer_z = data2;
//							observer_z = get_int16();
				current->observer_z = ((int) get_int16()) <<8;
				break;
			case SND_SET_O_Y:
				DEB1(fprt(stderr, "SND_O_Y\n"));
//ASYNC:					observer_y = data2;
//							observer_y = get_int16();
				current->observer_y = ((int) get_int16()) <<8;
				break;
			case SND_SET_O_ANGLE:
				DEB1(fprt(stderr, "SND_O_ANGLE\n"));
//ASYNC:					observer_angle = data2;
				current->observer_angle = (int) (int8) get_byte();
				break;

			case SND_SET_O_DX:
				DEB1(fprt(stderr, "SND_O_DX\n"));
//ASYNC:					observer_dx = data2;
//							observer_dx = (int) (int8) get_byte();
				current->observer_dx = (int) get_int16();
				break;
			case SND_SET_O_DZ:
				DEB1(fprt(stderr, "SND_O_DZ\n"));
//ASYNC:					observer_dz = data2;
//							observer_dz = (int) (int8) get_byte();
				current->observer_dz = (int) get_int16();
				break;
			case SND_SET_O_DY:
				DEB1(fprt(stderr, "SND_O_DY\n"));
//ASYNC:					observer_dy = data2;
//							observer_dy = (int) (int8) get_byte();
				current->observer_dy = (int) get_int16();
				break;

			case SND_O_DOPPLER_ON:
				DEB1(fprt(stderr, "SND_O_DOPPLER\n"));
//ASYNC:					observer_doppler = data2;
//							observer_flags |=SF_DOPPLER;
				break;
			case SND_O_DOPPLER_OFF:
				DEB1(fprt(stderr, "SND_O_DOPPLER_OFF\n"));
//ASYNC:					observer_doppler = 0;
//							observer_doppler = 0;
				break;
			default:
				fprintf(stderr, "Bad GSI command %d\n", command);
			} //switch (command2)
		}
	} //switch (command)
}


/*---------------------------------------------------------------------------*
 *  Initialize a sample (set conversion factors for speed scaling)
 *---------------------------------------------------------------------------*/
void init_sample(int slot, int speed, int type, int channels, int size) {
	int  i;

	current->sound_type[slot] = type;
	current->sound_speed[slot] = speed;
	current->sound_channels[slot] = channels;
	current->sound_size[slot] = size;
	DEB1(fprt(stderr, "  GSI_SERVER: set %d to: type=%d, speed=%d, channels=%d (% d bytes)\n", slot, type, speed, channels, size));
	if (current->sound_type[slot]==GSI_8BIT) {
		/* unsigned 8 bit to signed 8 bit */
		for(i=0; i<size; i++) current->sound_buffer[slot][i] ^=0x80;
	}

	// Conversion factors ..
	set_sound_conversion_of(slot);
}


/*---------------------------------------------------------------------------*
 *  Calulate speed conversion factors
 *  -- This stuff must be replaced with a run length algorithm
 *---------------------------------------------------------------------------*/
void set_sound_conversion() {
	int i;
	DEB1(fprt(stderr, "GSI_SERVER: set soundconversion\n"));

	for (i=0; i<num_clients; i++) {
		struct client_data *current = client_list[i];

		for (i=0; i< current->num_sounds; i++) {
			set_sound_conversion_of(i);
		}
	}
	//stereo -> mono: sound_conversion[i] *=2;
	//mono -> stereo: sound_conversion[i] /=2;
	//8bit -> 16 bit: H -> H,0
	//16bit -> 8 bit: H,L -> H
}

void set_sound_conversion_of(int slot) {
	if (current->sound_size[slot]>0) {
		current->sound_conversion[slot] = (current->sound_speed[slot]<<8) / dsp_speed;
		current->sound_conversion2[slot] = (dsp_speed<<8) / current->sound_speed[slot];
	}
}


/*---------------------------------------------------------------------------*
 *  Make a integer sinus table, values are 65536*sinus(2 pi* i/1024)
 *---------------------------------------------------------------------------*/
void make_sinus() {
	int  i;

	for (i=0; i<1024; i++) {
		sinus_tabel[i]= (int) (sin((M_PI*2.0*i)/1024.0) *65536.0 +.5);
	}
}


/*---------------------------------------------------------------------------*
 *  Pipe/socket reading functions:
 *     get_block, get_byte, get_int16, get_int32, get_string
 *---------------------------------------------------------------------------*/
void get_block(byte *data, int len) {
	fd_set read_fd_sel;
	int i=0, j=0;

	DEB1(fprt(stderr, "GSI_SERVER: sound_block: receiving a block of %d bytes\n", len));
	while((j=read (/*STDIN_FILENO*/ current->socket_fd, data+i, len-i)) <len-i) {
		if (j>0) i +=j;
		FD_ZERO(&read_fd_sel);
		FD_SET(/*STDIN_FILENO*/ current->socket_fd, &read_fd_sel);
		select(/*STDIN_FILENO*/ current->socket_fd+1, &read_fd_sel, NULL, NULL, NULL);
	}
}
byte get_byte() {
	fd_set read_fd_sel;
	byte pipe_b;
	while ( read(/*STDIN_FILENO*/ current->socket_fd, &pipe_b, sizeof(byte))<1 ) {
		FD_ZERO(&read_fd_sel);
		FD_SET(/*STDIN_FILENO*/ current->socket_fd, &read_fd_sel);
		select(/*STDIN_FILENO*/ current->socket_fd+1, &read_fd_sel, NULL, NULL, NULL);
	}

	return pipe_b;
}
int16 get_int16() { //get little endian int16 (change to MSB)
	byte b1, b2;

	b1=get_byte();
	b2=get_byte();
//	return b1 | (b2<<8); //LSB
	return b2 | (b1<<8); //MSB
}
int32 get_int32() { //get little endian int32 (change to MSB)
	byte b1,b2,b3,b4;

	b1=get_byte();
	b2=get_byte();
	b3=get_byte();
	b4=get_byte();
//	return b1 | (b2<<8) | (b3<<16) | (b4<<24); //LSB
	return b4 | (b3<<8) | (b2<<16) | (b1<<24); //MSB
}
char *get_string() {
	int  i, retval;
	int  size=256;
	char *tmp_string;
	char string_c;
	int  string_len=0;
	char *tmp_2;

	tmp_string=(char *) malloc(size);

	for (;;) {
		fd_set read_fd;
		FD_ZERO(&read_fd);
		FD_SET(/*STDIN_FILENO*/ current->socket_fd, &read_fd);
		retval = select(/*STDIN_FILENO*/ current->socket_fd+1, &read_fd, NULL, NULL, NULL);
		i = read(/*STDIN_FILENO*/ current->socket_fd, &string_c, sizeof(string_c));
		if ( i>0 ) {
			tmp_string[string_len++]=string_c;
			if (string_c==0) break;
			if (string_len==size) {
				size +=32; tmp_2=(char *) malloc(size);
				strncpy(tmp_2, tmp_string, size-32); 
				free(tmp_string);
				tmp_string=tmp_2;
			}
		}
	}
	return tmp_string;
}


/*---------------------------------------------------------------------------*
 *  Status information and callbacks
 *---------------------------------------------------------------------------*/
void status_send_byte(int data) {
	byte b=data;

	while (current->status_queue_len>=MAX_STATUS_QUEUE_LEN) status_flush();
	current->status_queue[current->status_queue_len++]=b;
}

// soft flush (flush as many bytes as possible without blocking)
void status_flush() {
	int  i,j;
	DEB1(fprt(stderr, "status flush\n"));
	for (i=0; i<num_clients; i++) {
		/*struct client_data * */current = client_list[i];

		if (current->status_queue_len >0 ) {
			DEB1(fprt(stderr, "GSI_SERVER: status flush of client %d (queue len=%d)\n", i, current->status_queue_len));
			DEB1(fprintf(stderr, "  -> status, writing %d bytes\n", current->status_queue_len));
			j=write (/*STDOUT_FILENO*/ current->socket_fd, current->status_queue, current->status_queue_len);
			if (j>0) {
				current->status_queue_len -=j;
				if (current->status_queue_len) memmove(current->status_queue, current->status_queue+j, current->status_queue_len);
			}
		}
		DEB1(fprt(stderr, "     client %d done (queue len=%d)\n", i, current->status_queue_len));
	}
}


/*---------------------------------------------------------------------------*
 *  Signal handlers
 *    quit gets called by SIGTERM, segfault by SIGSEGV or SIGFPE (for debug)
 *---------------------------------------------------------------------------*/
void quit (int i) {
	DEB1(fprt(stderr, "quit %s\n", server_name));
	exit (0);
}
#ifdef SEGFAULT_HANDLER
void segfault(int msg) {
	char *msgtext="";

	if (msg==SIGFPE) msgtext="SIGFPE";
	else if (msg==SIGSEGV) msgtext="SIGSEGV";
	fprintf(stderr, "\n%s: %s\n", server_name, msgtext);

	exit(-1);
}
#endif



/*---------------------------------------------------------------------------*
 *  Note on mixing: 
 *  What about overflow? Isn't it more natural for a fixed range system to use:
 *  1. Log's so that 255+255=255, 10+10 approx 20, 10+250=252 etc.
 *     This needs at least 2 256 b tables:
 *	     premix[j]+=logtable[*(sample[+j)];
 *	 Then after all adds:
 *	     premix[j]=premix[j]/num_playing;
 *	     premix[j]=exptable[premix[j];
 *  2. Fast approximation: just use max: loudness=max(loudness1, ..., loudness n)
 *	     premix[j]=((*(sample+j)) >premix[j] ? *(sample+j) : premix[j])
 *---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*
 *	EOF gsi_server.c
 *---------------------------------------------------------------------------*/
